/*
**	Nucleo para Multitasker
**
**	Rildo Pragana	02/21/91
**	Versao 1.0
*/

#define TID running->termid
#define DEL '\177'

#include "stdio.h"
#include "stdlib.h"
#include "stdarg.h"
#include "ctype.h"
#include "string.h"
#include "memory.h"
#include "dos.h"

#include "mtnet.h"
#include "task.h"
#include "mccntrl.h"

struct registers sched_sv[1]={0}; /* contexto salvo para retorno ao scheduler */

PROCESS *ready=NULL;
PROCESS *running=NULL;

SEMAF sdsp[NUMTERM]={0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,};
SEMAF sprt[NUMTERM]={0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,

		     0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,};

struct time_queue{
	struct time_queue *next;
	unsigned int delay;
	PROCESS *p;
};
static struct time_queue *wait_ticks_queue=NULL;
static byte timer_lock=0;

extern PROCESS task[];	/* arrays of processes */
extern void *init_tbl;	/* points to jobs descriptors table */

void *syslist[] = { (void *)&ready, (void *)&running, 
			(void *)sdsp, (void *)sprt, 
			(void *)sched_sv, (void *)&wait_ticks_queue,
			(void *)&init_tbl, (void *)task }; 

int access_num( ); /* em assembler no modulo swdos.asm */
void enqueue( PROCESS *, PROCESS ** );
PROCESS *dequeue( PROCESS ** );

int check_ready( ) {
	if (ready==NULL) return 0;
	return 1;
}

void scheduler( )
{
int r;
	/*return;*/ 		/* comentar p/rodar processos */
	if (ready==NULL) return;
	disable();
	while( ready!=NULL ) {
		running=dequeue(&ready);
		r=save_context(sched_sv);
		if (r==0) {
			enable();
			/* testa estouro de pilha 
			if (running->context->sp < 0x200 ||
			    running->context->sp > 0xf000) {
				mprintf("\f** STACK FULL **");
			}
			else */
				restore_context(running->context);
		}
	}
}

void wait( s )
SEMAF *s;
{
	int r;
	disable();
	enqueue(running,&(s->p));
	r=save_context(running->context);
	if (r==0)
	   restore_context(sched_sv); /* saida somente apos setjmp */
}

int _signal( s ) /* retorna 0 se nenhum task foi encontrado */
SEMAF *s;
{
	PROCESS *p;
	disable();
	if (s->p != NULL) {
		p=dequeue(&(s->p));
		enqueue( p, &ready );
		return 1;
	}
	return 0;
}

void enqueue( p, fila )
PROCESS *p;			     /* processo corrente */
PROCESS **fila; 		     /* fila de processos no semaforo */
{
	PROCESS *q;
	p->next=NULL;
	if (*fila==NULL) *fila=p;
	else {
		q=*fila;
		while (q->next!=NULL) q=q->next;
		q->next=p;
	}
}

PROCESS *dequeue( fila )
PROCESS **fila; 		  /* fila de processos no semaforo */
{
	PROCESS *p;
	p=*fila;
	if ( *fila != NULL )
		*fila=p->next;
	return p;
}

void start_task( p, stacklen, start, name, tid )
PROCESS *p;
int stacklen;
void (*start)();
char *name;
char tid;
{
	unsigned *stk;
	stk=(unsigned *)calloc(stacklen,sizeof(unsigned int));
	if (stk==NULL){
		puts("\nNo memory available to alloc stack");
		getchar();
		exit(1);
	}

	stk=stk+stacklen;
	save_context( p->context );
	p->base =
	p->context->sp=p->context->bp=FP_OFF(stk);
	p->context->ss=FP_SEG(stk);
	p->context->ip=FP_OFF(start);
	p->context->cs=FP_SEG(start);
	p->context->ds=		 /* segmento de dados de qualquer variavel */
	p->context->es=FP_SEG(syslist[0]); 
	p->name=name;
	p->termid=tid;
/*printf("\nTask stack: %04X:%04X",FP_SEG(stk),FP_OFF(stk));
printf("\nStart address: %04X:%04X",FP_SEG(start),FP_OFF(start));
printf("\nData segment: %04X\n",FP_SEG(p));
getchar();*/
	enqueue(p,&ready);
}

void signal_timer( ) {
	if (!timer_lock && (wait_ticks_queue != NULL))
		if (!(--wait_ticks_queue->delay)) {
			enqueue(wait_ticks_queue->p,&ready);
			wait_ticks_queue = wait_ticks_queue->next;
		}
}

void ticks_delay( dly ) 
unsigned int dly;
{
	int r;
	struct time_queue tmp,*tq;
	tmp.p=running;
	timer_lock=1;
	if (wait_ticks_queue==NULL) {
		tmp.delay=dly;
		tmp.next=NULL;
		wait_ticks_queue=&tmp;
		disable();
		timer_lock=0;
		r=save_context(running->context);
		if (r==0)
			restore_context(sched_sv);
		if (wait_ticks_queue != NULL)
			if (!wait_ticks_queue->delay) {
				enqueue(wait_ticks_queue->p,&ready);
				wait_ticks_queue = wait_ticks_queue->next;
			}
		return;
	}
	tq=wait_ticks_queue;
	if (dly<tq->delay) {
		tmp.delay=dly;
		tmp.next=wait_ticks_queue;
		wait_ticks_queue=&tmp;
		tq->delay -= dly;
		disable();
		timer_lock=0;
		r=save_context(running->context);
		if (r==0)
			restore_context(sched_sv);
		if (wait_ticks_queue != NULL)
			if (!wait_ticks_queue->delay) {
				enqueue(wait_ticks_queue->p,&ready);
				wait_ticks_queue = wait_ticks_queue->next;
			}
		return;
	}
	dly -= tq->delay;
	while (tq->next!=NULL) {
		if (dly>=tq->next->delay) {
			dly -= tq->next->delay;
			tq=tq->next;
		}
		else {
			tmp.delay=dly;
			tmp.next=tq->next;
			tq->next=&tmp;
			tq->delay -= dly;
			disable();
			timer_lock=0;
			r=save_context(running->context);
			if (r==0)
				restore_context(sched_sv);
			if (wait_ticks_queue != NULL)
				if (!wait_ticks_queue->delay) {
					enqueue(wait_ticks_queue->p,&ready);
					wait_ticks_queue = wait_ticks_queue->next;
				}
			return;
		}
	}
	tmp.delay=dly;
	tmp.next=NULL;
	tq->next=&tmp;
	disable();
	timer_lock=0;
	r=save_context(running->context);
	if (r==0)
		restore_context(sched_sv);
	if (wait_ticks_queue != NULL)
		if (!wait_ticks_queue->delay) {
			enqueue(wait_ticks_queue->p,&ready);
			wait_ticks_queue = wait_ticks_queue->next;
		}
}
