/*
**	Rotinas auxiliares do MCobol Debugger
**
**	Rildo Pragana -- 5/14/92
*/

#include <stdio.h>
#include <dos.h>

struct registers {
	unsigned int cs,ip,bx,cx,dx,bp,ds,es,ss,sp,si,di;
};

struct process_type {
	char *name;			/* ptr p/nome do processo */
	char termid;			/* endereco do mt100 */
	struct registers context[1];	/* area p/salvar o contexto */
	struct process_type *next;	/* link p/proximo da fila */
	unsigned int base;		/* init value for bp */
};
typedef struct process_type PROCESS;

typedef struct {
	int val;			/* valor do semaforo */
	PROCESS *p;			/* fila de processos em wait */
} SEMAF;

void wait( SEMAF * );
int _signal( SEMAF * );
static void enqueue( PROCESS *p, PROCESS **fila );
static PROCESS *dequeue( PROCESS **fila );
void interrupt (*_getvect(int intn))();
void _setvect(int intn, void interrupt (*isr)());
void _intr( int intn, struct REGPACK *regs );

/*
void *syslist[] = { (void *)&ready, (void *)&running,
			(void *)sdsp, (void *)sprt,
			(void *)sched_sv, (void *)&wait_ticks_queue,
			(void *)&init_tbl, (void *)task };
*/

PROCESS **ready;		/* syslist+0 */
PROCESS **running;		/* syslist+1 */
struct registers *sched_sv;	/* syslist+4 */

#define MAXBKP 20
SEMAF bkp_stop;
static struct breakpoint {
	unsigned char *p;
	unsigned line;
	unsigned char c;
} bkp[MAXBKP] =
	{{NULL,0,0xcc},{NULL,0,0xcc},{NULL,0,0xcc},{NULL,0,0xcc},{NULL,0,0xcc},
	 {NULL,0,0xcc},{NULL,0,0xcc},{NULL,0,0xcc},{NULL,0,0xcc},{NULL,0,0xcc},
	 {NULL,0,0xcc},{NULL,0,0xcc},{NULL,0,0xcc},{NULL,0,0xcc},{NULL,0,0xcc},
	 {NULL,0,0xcc},{NULL,0,0xcc},{NULL,0,0xcc},{NULL,0,0xcc},{NULL,0,0xcc}};
unsigned char bkpw=0;

struct ilist {
	unsigned char functor;
	int value;
	struct ilist *next;
};

/******** funcoes da biblioteca do Turbo Prolog (alocacao dinamica) *******/
void *_malloc(unsigned size);
void *_free(void *);
void *alloc_gstack(unsigned);
void console_cursor( unsigned char *,unsigned char * );
void console_gotoxy( unsigned char,unsigned char );
void console_clrscr( void );
void console_write( unsigned char );

static void interrupt (*Int03Old)();
void interrupt Int03( void );		/* em assembler */

/*void interrupt Int03( bp,di,si,ds,es,dx,cx,bx,ax,ip,cs,flags )
unsigned bp,di,si,ds,es,dx,cx,bx,ax,ip,cs,flags;*/

/*
static void interrupt (*Int10Old)();
void interrupt Int10( void );		/* em assembler */
union reg2 {
	unsigned x;
	struct {
		unsigned char l;
		unsigned char h;
	} h;
};

unsigned proc_Int10( func, par )
unsigned func;
union reg2 par;
{
	union reg2 res;
	switch( func>>8 ) {
	case 2:
		console_cursor( &(res.h.l),&(res.h.h) );
		break;
	case 3:
		console_gotoxy( par.h.l,par.h.h );
		break;
	case 9:
		console_clrscr();
		break;
	case 0x0e:
		console_write( func & 0x0ff );
		break;
	}
	return res.x;
}
*/

void proc_Int03( unsigned cs, unsigned ip ) {
	int i;
	unsigned char *currip;
	currip = MK_FP( cs,ip );
	for (i=0;i<MAXBKP;i++) {
		if (bkp[i].p == currip) {
			*(bkp[i].p) = bkp[i].c;
			bkp[i].c = 0xcc; /* indica bkpc disponivel */
			break;
		}
	}
	/* ip--; 		<-- executado em assembly */
	wait(&bkp_stop);
	bkpw=0;
}

void wait( s )
SEMAF *s;
{
	int r;
	disable();
	enqueue(*running,&(s->p));
	r=save_context((*running)->context);
	if (r==0)
	   restore_context(sched_sv); /* saida somente apos setjmp */
}

int _signal( s ) /* retorna 0 se nenhum task foi encontrado */
SEMAF *s;
{
	PROCESS *p;
	disable();
	if (s->p != NULL) {
		p=dequeue(&(s->p));
		enqueue( p, ready );
		return 1;
	}
	return 0;
}

void enqueue( p, fila )
PROCESS *p;			     /* processo corrente */
PROCESS **fila; 		     /* fila de processos no semaforo */
{
	PROCESS *q;
	p->next=NULL;
	if (*fila==NULL) *fila=p;
	else {
		q=*fila;
		while (q->next!=NULL) q=q->next;
		q->next=p;
	}
}

PROCESS *dequeue( fila )
PROCESS **fila; 		  /* fila de processos no semaforo */
{
	PROCESS *p;
	p=*fila;
	if ( *fila != NULL )
		*fila=p->next;
	return p;
}

void install_trap_0( int sw )
{
	if (sw) {
		struct REGPACK r;
		r.r_ax = 0xef04;
		_intr( 0x2f,&r );
		ready = *((PROCESS ***)MK_FP(r.r_ds,r.r_si));
		running  = *((PROCESS ***)MK_FP(r.r_ds,r.r_si+4));
		sched_sv = *((struct registers **)MK_FP(r.r_ds,r.r_si+16));
		Int03Old = _getvect( 3 );
		_setvect( 3,Int03 );
/*		Int10Old = _getvect( 10 );
		_setvect( 10,Int10 );*/
	}
	else {
		_setvect( 3,Int03Old );
/*		_setvect( 10,Int10Old );*/
	}
}

/********* instala breakpoint *************/

void breakpt_0( unsigned Seg, unsigned Off, unsigned line ) {
	int i;
	unsigned char *tmp;
	tmp = MK_FP( Seg,Off );
	for (i=0;i<MAXBKP;i++) {
		if (bkp[i].c==0xcc) { /* so' instala se bkpc disponivel */
			bkp[i].p = tmp;
			bkp[i].c = *(bkp[i].p);
			*(bkp[i].p) = 0xcc;	/* INT 03 single byte code */
			bkp[i].line = line;
			break;
		}
	}
}

void clr_breakpt_0( unsigned Seg, unsigned Off ) {
	int i;
	unsigned char *currip;
	currip = MK_FP( Seg,Off );
	for (i=0;i<MAXBKP;i++) {
		if ((bkp[i].p == currip) &&
		    (bkp[i].c != 0xcc)) {
			*(bkp[i].p) = bkp[i].c;
			bkp[i].c = 0xcc; /* indica bkpc disponivel */
			break;
		}
	}
}

void toggle_breakpt_0( unsigned Seg, unsigned Off, unsigned line ) {
	int i;
	unsigned char *currip;
	currip = MK_FP( Seg,Off );
	for (i=0;i<MAXBKP;i++) {
		if (bkp[i].p == currip) {
			if (bkp[i].c==0xcc) {
				bkp[i].c = *(bkp[i].p);
				*(bkp[i].p) = 0xcc;
				bkp[i].line = line;
				return;
			}
			else {
				*(bkp[i].p) = bkp[i].c;
				bkp[i].c = 0xcc;
				return;
			}
		}
	}
	for (i=0;i<MAXBKP;i++) {
		if (bkp[i].c==0xcc) { /* so' instala se bkpc disponivel */
			bkp[i].p = currip;
			bkp[i].c = *(bkp[i].p);
			*(bkp[i].p) = 0xcc;
			bkp[i].line = line;
			return;
		}
	}
}

void list_breakpt_0( struct ilist **list )
{
	int i;
	struct ilist *p;
	for (i=0;i<MAXBKP;i++) {
		if (bkp[i].c != 0xcc) {
			p = *list = alloc_gstack(sizeof(struct ilist));
			p->functor = 1;  /* active list element */
			p->value = bkp[i].line;
			list = &(*list)->next;
		}
	}
	p = *list = alloc_gstack(sizeof(char));
	p->functor = 2;	/* end of list */
}

void wait_breakpt_0( void ) {
	enable();
	while(bkpw) ;
}

void crestart_0( void ) {
	if (_signal(&bkp_stop))
		bkpw++;
}

void frontlabel_0( char *s, char **f, char **r ) {
	int i;
	char *tmp;
	while (*s && (*s == ' ')) s++;
	tmp=s;
	for ( i=0 ; *tmp && (*tmp != ' ') ; i++,tmp++ ) ;
	*f = tmp = alloc_gstack( i+1 );
	while (	(*tmp = *s) && (*s != ' ') ) { tmp++; s++; }
	*tmp = 0;
	tmp=s;
	for ( i=0 ; *tmp ; i++ ) tmp++;
	*r = tmp = alloc_gstack( i+1 );
	while (*tmp++ = *s++) ;
	*tmp = 0;
}

void cget_string_0( unsigned Seg, unsigned Off, unsigned width, char **str )
{
	char *t,*s;
	unsigned i;
	t = (char *)MK_FP( Seg,Off );
	*str = s = alloc_gstack( width+1 );
	for (i=0;i<width;i++)
		*s++ = *t++;
	*s = 0;
}

void cget_string1_0( unsigned Seg, unsigned Off, unsigned width, char **str )
{
	char *t,*s;
	unsigned i;
	t = (char *)MK_FP( Seg,Off );
	*str = s = alloc_gstack( width*2+1 );
	width--;
	for (i=0;i<width;i++)  {
		*s++ = ((*t >> 4) & 0x0f) + '0';
		*s++ = (*t++ & 0x0f) + '0';
	}
	*s++ = ((*t >> 4) & 0x0f) +  '0';
	*s++ = ((*t & 0x0f)==0x0c) ? '+' : '-';
	*s = 0;
}

/* end of MCDEBC.C */
