/*
**	     BPLUS file indexing program - Version 1.1
**
**	Adaptado para uso com o HandTerm
**	R.Pragana -- 1993
*/


#define IX_OK       1
#define IX_FAIL     0
#define EOIX       (-2)
#define MAXKEY      100   /* chave maxima permitida */
#define NUM_BUFS    8
#define MAX_LEVELS  8
#define IXB_SIZE    512
#define IXB_SPACE  (IXB_SIZE - sizeof(int) - sizeof(long) * 2)

typedef long RECPOS;

typedef struct                    /*  entry structure in index        */
  {  RECPOS   idxptr;             /*  points to lower index level     */
     RECPOS   recptr;             /*  points to data record           */
     char     key[MAXKEY];        /*  start of record key             */
  }  ENTRY;

typedef struct                    /*  index record format             */
  {  RECPOS   brec;               /*  position in index file          */
                                  /*  or location of next free block  */
     int      bend;               /*  first unused block location     */
     RECPOS   p0;                 /*  points to next level            */
     char     entries[IXB_SPACE]; /*  here are the key entries        */
  }  BLOCK;

typedef struct                    /*  disk file info                  */
  {  RECPOS   ff;                 /*  location of first free block    */
     int      nl;                 /*  number of index levels          */
  }  IX_DISK;

typedef struct                    /*  memory buffer pool of indx blks */
  {  int      dirty;              /*  true if changed                 */
     int      handle;             /*  index file handle               */
     int      count;              /*  number of times referenced      */
     BLOCK    mb;
  }  MEMBLOCK;

typedef struct
  {  MEMBLOCK     cache [ NUM_BUFS ];
  }  IX_BUFFER;

typedef struct                    /*  in-memory index descriptor      */
  {  int      ixfile;		  /* file handle		      */
     int      level;              /*  level in btree                  */
     int      duplicate;          /*  no duplicate keys if 0          */
     struct
       {  RECPOS    cblock;       /*  position in index file          */
          int       coffset;      /*  current offset within block     */
       }  pos [ MAX_LEVELS ];
     BLOCK    root;               /*  root index record               */
     IX_DISK  dx;
  }  IX_DESC;

int  open_index(char *,IX_DESC *, int); /* abre arquivo de indice */
int  close_index(IX_DESC *);		/* fecha indice */
int  make_index(char *,IX_DESC *, int); /* cria novo arquivo de indice */
int  first_key(IX_DESC *);		/* posiciona na primeira chave */
int  last_key(IX_DESC *);		/* posiciona na ultima chave */
int  next_key(ENTRY *, IX_DESC *);	/* avanca para a proxima chave
					   (ordem alfabetica) */
int  prev_key(ENTRY *, IX_DESC *);	/* recua p/chave anterior
					   (ordem alfabetica) */
int  find_key(ENTRY *, IX_DESC *);	/* procura por uma determinada
					   chave, colocada em ENTRY *
					   no campo key[] */
int  add_key(ENTRY *, IX_DESC *);	/* adiciona no indice, usando
					   chave em (ENTRY *)-> key
int  locate_key(ENTRY *, IX_DESC *);	/* semelhante ao find_key
					   porem localizando a primeira
					   chave maior ou igual `a dada */
int  delete_key(ENTRY *, IX_DESC *);	/* retira uma chave dada do 
					   arquivo de indice */
int  find_exact(ENTRY *, IX_DESC *);	
