/*
** 	Rotinas de comunicao mestre/escravo
** 	Micro Cobol Runtime Library
**
** 	Rildo Pragana -- 06/92
**
*/
#include <stdio.h>
#include <dos.h>

#include "mtnet.h"
#include "task.h"
#include "mccntrl.h"

void send_transaction( struct  mailbox_type * );

/*******  COM port addresses & bits definitions ********/
#define COMBASE 0x3f8
#define COMPORT COMBASE
#define RXCHAR() (inportb( COMPORT ))
#define TXCHAR(c) (outportb( COMPORT,c ))
#define COMIER  (COMBASE+1)	/* interrupt enable register */
/* 0 | 0 | 0 | 0 | Modem st | Receive line st | Tx empty | Rx available */
#define COMIIR  (COMBASE+2)	/* interrupt identification register */
/* 0 | 0 | 0 | 0 | 0 | interrupt ID (bits 2-0):
			(bit 0 = 0 means interrupt pending)
			110 - line status
			100 - receive data
			010 - tx buffer empty
			000 - modem status */
#define COMLCR	(COMBASE+3)	/* line control register */
/* divisor latch access bit | break | bits 5-3
					000=no parity
					001=odd parity
					011=even parity
						| bit 2
						  0=1 stop bit
						  1=2 stop bits
						  | bits 1-0
						    10 - 7 bits/char
						    11 - 8 bits/char */
#define COMMCR  (COMBASE+4)	/* modem control register */
/* 0 | 0 | 0 | loopback test | OUT2 | OUT1 | RTS | DTR */
#define RTSON() outportb( COMMCR,0x03 )
#define RTSOFF() outportb( COMMCR,0x01 )
#define COMLSR  (COMBASE+5)	/* line status register */
/* 0 | tx empty | tx holding reg empty | Break detected |
	framing error | parity error | overrun error | rx data ready */
#define WAITRX() while( !(inportb( COMLSR ) & 0x01) )
#define WAITTX() while( !(inportb( COMLSR ) & 0x20) )
#define WAITTXE() while( (inportb( COMLSR ) & 0x40) )
#define COMMSR	(COMBASE+6)	/* modem status register */
/* RLSD | RI | DSR | CTS | Delta RLSD | Delta RI | Delta DSR | Delta CTS */
#define CTSTEST() (inportb( COMMSR ) & 0x10)
#define SEND55	WAITTX(); TXCHAR(0); WAITTX(); TXCHAR(0); WAITTX(); TXCHAR(0x55);
#define WAIT55	b=0; while (b!=0x55) { WAITRX(); b = RXCHAR(); }


void set_baudrate( divisor )
unsigned divisor;
/********** divisor = 1843200 / (16 * baudrate ) **********/
{
	unsigned char oldval;
	disable();
	oldval = (unsigned char)inportb(COMLCR);
	outportb( COMLCR, (0x80 | oldval) );
	outportb( COMPORT, (divisor & 0xff) );
	outportb( COMIER, (divisor >> 8) );
	outportb( COMLCR, oldval );
	enable();
}

/*
**	INITCOMM deve ser chamada pelo processo dummy
**	uma unica vez, antes de permanecer em loop
**	aguardando transacoes a enviar.
**	No master, deve ser chamada inicialmente (tambem uma so' vez)
**	antes de permanecer em loop aguardando transacoes pela linha.
*/
void INITCOMM( )
{
	union REGS xr,yr;
	xr.h.ah=0;
	xr.h.al=0xe3; /* baud=9600, parity=none, stop=1, wordlength=8 */
	xr.x.dx=0;	/* portnumber 0 (COM1) */
	int86(0x14,&xr,&yr);
	set_baudrate(1);		/* 115200 bps */
	RTSOFF();
}

static SEMAF dummy_sleep={0,NULL};
static struct mailbox_type {
	unsigned char *tran;
	int tranlen;
	unsigned char *buf;
	int buflen;
	SEMAF in_process;
	struct mailbox_type *next;
};

static struct mailbox_type *dummy_mailbox=NULL;

/*
**	REMOTETRAN e' chamada por cada processo que precisa
**	enviar uma transacao ao master e obter uma resposta (obrigatoria).
**	O processo fica suspenso ate' o master processar completamente
**	a transacao e devolver pela linha serial a resposta.
**	TRAN e' a transacao a solicitar, TLEN e' o seu tamanho
**	BUF e' o buffer para a resposta, BLEN o seu tamanho
**	E' importante que o programa do master analise o tipo de transacao
**	e concorde com os tamanhos, pois nao e' feito um check
**	do tamanho dos buffers/transacoes enviados ou recebidos.
**	Assim, por exemplo, um campo (que pode ser de 1 caracter) indica
**	o tipo de transacao, associado a um tamanho pre-determinado.
*/
void REMOTETRAN( TRAN,TLEN,BUF,BLEN )
unsigned char *TRAN;
char *TLEN;
unsigned char *BUF;
char *BLEN;
{
	struct mailbox_type mailbox,*tmp;
	mailbox.tran = TRAN;
	mailbox.buf = BUF;
	mailbox.tranlen = (*TLEN - '0')*100 + (*(TLEN+1) - '0')*10 +
			*(TLEN+2) - '0';
	mailbox.buflen = (*BLEN - '0')*100 + (*(BLEN+1) - '0')*10 +
			*(BLEN+2) - '0';
	mailbox.in_process.p = NULL;
	mailbox.next=NULL;
	tmp=dummy_mailbox;
	if (tmp==NULL)
		dummy_mailbox=&mailbox;
	else {
		while (tmp->next != NULL) tmp=tmp->next;
		tmp->next=&mailbox;
	}
	_signal(&dummy_sleep);
	wait(&(mailbox.in_process));
}

/*
**	SENDTRAN e' usada por um processo dummy para enviar/receber
**	transacoes pela linha serial (nos slaves, e nao no master)
**	O processo dummy contem apenas um loop, chamando esta rotina.
**	Opcionalmente, podemos utilizar este processo para monitorar
**	os pedidos efetuados nos microcomputadores remotos.
*/
void SENDTRAN( )
{
	int cnt;
	struct mailbox_type *mailbox;
	if (dummy_mailbox==NULL)
		wait(&dummy_sleep);
	mailbox = dummy_mailbox;
	dummy_mailbox = dummy_mailbox->next;
	send_transaction( mailbox );
	_signal(&(mailbox->in_process));
}

void send_transaction( mailbox )
struct mailbox_type *mailbox;
{
	unsigned char b;
	/* primeiro liga RTS */
	RTSON();
	/* agora aguarda o CTS do master */
	while (!CTSTEST()) ;
	disable();
	/* envia tamanho da transacao p/o master (2 bytes) */
	b = (unsigned char ) (mailbox->tranlen & 0xff);
	WAITTX(); TXCHAR( b );
	b = (unsigned char) (mailbox->tranlen >> 8);
	WAITTX(); TXCHAR( b );
	/* envia transacao para o master */
	b=16;
	while (mailbox->tranlen--) {
		WAITTX(); TXCHAR( *(mailbox->tran++) );
		if (!b--) {
			enable();
			WAITRX(); RXCHAR();
			disable();
			b=16;
		}
	}
	enable();
	while (CTSTEST()) ;
	RTSOFF();
	while (!CTSTEST()) ;
	disable();
	RTSON();
	/* aguarda resposta do master e coloca-a no buffer */
	b=16;
	while (mailbox->buflen--) {
		WAITRX(); *(mailbox->buf++) = RXCHAR();
		if (!b--) {
			enable();
			b=16;
			WAITTX();
			disable();
			TXCHAR(0);
		}
	}
	/* desliga RTS */
	RTSOFF();
	while (CTSTEST()) ;
	enable();
}

/*
**	MASTERTRAN recebe uma transacao para o master, de tamanho
**	determinado pelo "caller", com tipos declarados pelo
**	programador, preenchendo-a no buffer TRAN e retornando
**	ao programa MCobol.
**	O programa do master ficara' em loop ao chamar esta rotina
**	ate' que uma transacao seja recebida.
**	O parametro DELAY informa quanto deve ser esperado antes
**	de testar novamente o advento de outra transacao
*/
void MASTERTRAN( TRAN,DELAY )
unsigned char *TRAN;
char *DELAY;
{
	int ticks;
	unsigned len;
	unsigned char b;
	ticks = (*DELAY - '0') * 10 + *(DELAY+1) - '0';
	while (!CTSTEST()) /* RTS ainda nao ativado pelo remoto */
		ticks_delay(ticks);
	/* liga CTS */
	disable();
	RTSON(); while(!CTSTEST()) ;
	/* recebe tamanho da transacao como unsigned (2 bytes) */
	WAITRX(); len = RXCHAR();
	WAITRX(); len |= (RXCHAR() << 8);
	/* recebe transacao em TRAN */
	b=16;
	while(len--) {
		WAITRX(); *TRAN++ = RXCHAR();
		if (!b--) {
			enable();
			b=16;
			WAITTX();
			disable();
			TXCHAR(0);
		}
	}
	enable();
	RTSOFF();
	while (CTSTEST()) ;
}

/*
**	MASTERSEND retorna pela linha o conteudo de um buffer
**	de tamanho determinado pelo programador, que deve coincidir
**	com o buffer esperado pelo programa remoto.
*/
void MASTERSEND( BUF,LEN )
unsigned char *BUF;
char *LEN;
{
	unsigned char b;
	int cnt;
	cnt = (*LEN - '0') * 100 + (*(LEN+1) - '0') * 10 +
		*(LEN+2) - '0';
	disable();
	RTSON();
	while(!CTSTEST()) ;
	/* envia CNT bytes do buffer BUF pela linha */
	b=16;
	while(cnt--) {
		WAITTX();
		TXCHAR( *BUF++ );
		if (!b--) {
			enable();
			WAITRX(); RXCHAR();
			disable();
			b=16;
		}
	}
	/* desliga CTS */
	enable();
	while (CTSTEST()) ;
	RTSOFF();
}

/* end of COMMUN.C */
