/* GETARGS.C	Command line argument processor for C programs
**
** (C) 1985  Allen I. Holub - Dr. Dobb's Journal, May 1985 pg. 32
*/

#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "getargs.h"

typedef int (*PFI)();
static char *setarg( ARG *, char * );
static ARG *findarg( int, ARG *, int );
static void pr_usage( ARG *, int );
int getargs( int, char **, ARG *, int );
static int stoi( char ** );

static int stoi( s )
char **s; {
	register int num=0;
	register char *str;
	str=*s;
	while (*str==' '||*str=='\t'||*str=='\n') str++;
	while ('0'<=*str && *str<='9') {
		num*=10;
		num+=*str++-'0';
	}
	*s=str;
	return num;
}

static char *setarg( argp, linep )
ARG *argp;
char *linep;  {
	/* set an argument. argp points at the argument table entry
	   corresponding to *linep. Return linep, updated to point
	   past the argument being set */

	++linep;
	switch( argp->type )  {
	case INTEGER:
		*argp->variable = stoi( &linep );
		break;
	case BOOLEAN:
		*argp->variable = 1;
		break;
	case CHARACTER:
		*argp->variable = *linep++;
		break;
	case STRING:
		*(char **)argp->variable = linep;
		linep = "";
		break;
	case PROC:
		(* (PFI)(argp->variable))( linep );
		linep  =  "";
		break;
	default:
		printf("Erro: tipo de argumento invalido!\n");
		break;
	}
	return linep;
}

static ARG *findarg( c, tabp, tabsize )
int c,tabsize;
ARG *tabp; {
	/* Return pointer to argument table entry corresponding
	   to c (or NULL if c isn't in table). */
	for (;--tabsize>=0;tabp++)
		if (tabp->arg==c)
			return tabp;
	return NULL;
}

static void pr_usage( tabp, tabsize )
ARG *tabp;
int tabsize; {
	/* Print the argtab in the form:
		-<arg> <errmsg>  ( value is <*variable> ) */
	for (;--tabsize>=0;tabp++)  {
		switch( tabp->type ) {
		case INTEGER:
			printf("-%c<num> %-40s (com valor ",
				tabp->arg,tabp->errmsg);
			printf("%-5d)\n",*(tabp->variable));
			break;
		case BOOLEAN:
			printf("-%c    %-40s (com valor ",
				tabp->arg,tabp->errmsg);
			printf("%-5s)\n",*(tabp->variable)
				? "ON" : "OFF");
			break;
		case CHARACTER:
			printf("-%c<c>  %-40s (com valor ",
				tabp->arg,tabp->errmsg);
			printf("%-5c)\n",*(tabp->variable));
			break;
		case STRING:
			printf("-%c<str> %-40s (com valor ",
				tabp->arg,tabp->errmsg);
			printf("<%s>)\n",*(char **)tabp->variable);
			break;
		case PROC:
			printf("-%c<str>   %-40s\n",
				tabp->arg,tabp->errmsg);
			break;
		}
	}
}

#define ERRMSG "Argumento ilegal <%c>. Os argumentos legais sao:\n\n"

int getargs( argc, argv, tabp, tabsize )
int argc, tabsize;
char **argv;
ARG *tabp; {
	/* Process command line arguments. Stripping all comand line
	switches out of argv. Return a new argc. If an error is found
	exit(1) is called (getargs won't return) and a usage message
	is printed showing all arguments in the table. */

	register int nargc;
	register char **nargv, *p;
	register ARG *argp;
	nargc = 1;
	for (nargv = ++argv;--argc>0;argv++) {
		if (**argv != '-') {
			*nargv++ = *argv;
			nargc++;
		}
		else {
			p = (*argv)+1;
			while (*p) {
				if ( (argp=findarg(*p,tabp,tabsize))!=0)
					p = setarg(argp,p);
				else {
					printf(ERRMSG,*p);
					pr_usage(tabp,tabsize);
					exit(1);
				}
			}
		}
	}
	return nargc;
}
