/*
**  Cobol Compiler Library -- Accept functions module
**
**  by Rildo Pragana -- 1992
**  P.O. Box 7440 - Recife PE Brazil 50000
**
**  Revision history:
**	29-oct-91 Code implementation
**	24-dec-91 Separate library for files
**	27-apr-92 Separate file for accept functions
**		(except 'accept from inkey')
**	07-mai-92 Input to accept buffer
*/

#include "mcoblib.h"

char proc_accept( char,byte );
void msetpos( byte * );
void mgetpos( byte * );
void liga_alarmes( char * ) ;

/* constantes relacionadas com a programacao do display */
#define CLRSCR	1
#define LIN1	0x80
#define LIN2	0xc0
#define CURLEFT 04
#define CURRIG	06
#define CUROFF  0x0c
#define CURON   0x0e
#define CURDL   0x10
#define MAXCOLS 16

static struct accept_buf {
	struct fld_desc *f;
	char *buffer;
	char *buf;
	byte pos[2]; 	/* para armazenar a posicao apos a msg */
	byte cnt;	/* posicoes que faltam ser preenchidas */
	byte digcnt;	/* no. chars efetivamente digitados */
	byte echo;
	byte sign;
	byte decimal_flg;
	char ret;	/* retorno (c/sinal) */
	SEMAF filled;
} abuf[NUMTERM]={NULL};

extern PROCESS *running;
extern char tascii[4][16];
extern byte mcol[NUMTERM];
extern byte mlin[NUMTERM];

char teclas[NUMTERM][16];	/* buffer - 16 teclas por terminal */
byte tec_ent[NUMTERM]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
byte tec_sai[NUMTERM]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

SEMAF skey[NUMTERM]={0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,};

/*
**	Trata eventos da caneta otica
**	Versao DEMO -- somente um terminal por vez
**
**	NAO DEVE SER USADA COMO SUBROTINA
*/
int tabn;
int bdir;
char codigo[14];
char icod;
char code7;
char bins[14];
char binptr=0;

char bcodea[] = {0x0d,0x19,0x13,0x3d,0x23,0x31,0x2f,0x3b,0x37,0x0b};
char bcodeb[] = {0x27,0x33,0x1b,0x21,0x1d,0x39,0x05,0x11,0x09,0x17};
char bcodec[] = {0x39,0x33,0x1b,0x21,0x17,0x27,0x05,0x11,0x09,0x1d};
char bcodei[] = {0x0b,0x13,0x19,0x2f,0x31,0x23,0x3d,0x37,0x3b,0x0d};
char bcode7[] = {0x00,0x0b,0x0d,0x0e,0x13,0x19,0x1c,0x15,0x16,0x1a};

int validate_code() {
	int i;
	for (i=0;i<14;i++)
		if (codigo[i]==0) return 0;
	for (i=0;i<10;i++) {
		if (code7 == bcode7[i])
			return (i+'0');
	}
	return 0;
}

void bdecode( i )
int i;
{
	int x,c;
	c = 0;
	if ((tabn==7) || (tabn==14)) {
		if (i==5)
			codigo[icod++] = '*';
		return;
	}
	if (tabn==1) bdir=0; /* suponha direcao normal */
	for (x=0;x<10;x++) {
		if (tabn==1)
		{
			if (i==*(bcodea+x)) {
				c=x+'0';
				bdir=0;
				break;
			}
			if (i==*(bcodeb+x)) {
				c=x+'0';
				bdir=1;
				break;
			}
		}
		else {
			if ( ((tabn<7) && !bdir) ||
			     ((tabn>7) && bdir) )
			{
				if (i==*((bdir ? bcodec : bcodeb)+x)) {
					c=x+'0';
					code7 |= (0x40 >>
						(bdir ? 14-tabn : tabn));
					break;
				}
				if (i==*((bdir ? bcodei : bcodea)+x)) {
					c=x+'0';
					break;
				}
			}
			else
				if (i==*((bdir ? bcodeb : bcodec)+x)) {
					c=x+'0';
				break;
			}
		}
	}
	codigo[icod++] = c;
}

void bputc( term,codigo )
byte term;
char codigo;
{
	char ret;
	if (ret = proc_accept( codigo,term )) {
		abuf[term].f = NULL;
		abuf[term].ret = ret;
		_signal( &(abuf[term].filled) );
	}
}

void barcode_read( term )
byte term;
{
	int i,c;
	tabn = 1;
	icod = 0;
	code7 = 0;
	for (i=0;i<14;i++) {
		codigo[i]=0;
		bdecode(bins[i]);
		tabn++;
	}
	if (c=validate_code()) {
		bputc(term,c); /* poe primeiro codigo no buffer do terminal */
		if (bdir) {
			for (i=12;i>=0;i--)
				if (i!=6)
					bputc(term,codigo[i]);
		}
		else {
			for (i=0;i<13;i++)
				if (i!=6)
					bputc(term,codigo[i]);
		}
	}
	binptr = 0;
}

void process_lpen( term,c )
byte term;
byte c;
{
	bins[binptr++]=c;
	if (binptr==14) /* final do codigo */
		barcode_read(term);
}

/*
**	Processa eventos
**	Armazena ascii correspondente `a sequencia de teclas lida
** 	Pode ser modificada para definir outro layout de teclado.
**
**	NAO DEVE SER USADA COMO SUBROTINA EM PROCESSOS
*/
void process_event( term,ev )
byte term;
struct evento ev;
{
#define ESTADOS 5
#define EVENTOS 8

/*		estados possiveis:
**
**	SHIFT keys = [*], [0], [#]
**
**		1		estado inicial
**		2 x 3	estados de SHIFT		( make e break de SHIFT key )
**	3 x 3	estados de combinacao	( [SHIFT] + [k] :
**											[SHIFT A] & [SHIFT B]
**					eventos possiveis:
**
**	0 - 1,F1 	1 - 0,F1 	2 - 1,F2 	3 - 0,F2
**	4 - 1,F3 	5 - 0,F3 	6 - 1,n 	7 - 0,n
**	|   | |
**	|   | +---> tecla afetada
**	|   +-----> 1= make , 0= break
**	+---------> evento = classe
*/
	static struct transicao {
		byte ps;	/* proximo estado */
		char ascii;	/* codigo a retornar
					0 - nulo
					1 a 4 - tecla de entrada
					outros - ascii */
	} transicoes[ ESTADOS ] [ EVENTOS ] = {

/*							EVENTOS
** ------------------------------------------------------------------------
**	F1	-F1	F2	-F2	F3	-F3	n	-n
** ------------------------------------------------------------------------
*/
	1,0,	0,0,	2,0,	0,0,	3,0,	0,0,	4,1,	0,0,  /* ---> Estado inicial   */
	1,0,	0,0,	2,0,	0,0,	3,0,	0,0,	1,2,	1,0,  /* -+		       */
	1,0,	0,0,	2,0,	0,0,	3,0, 	0,0,	2,3,	2,0,  /*  |-> Estados de SHIFT */
	1,0,	0,0,	2,0,	0,0,	3,0,	0,0,	3,4,	3,0,  /* -+		       */
	1,0,	0,0,	2,0,	0,0,	3,0,	0,0,	0,0,	0,0   /* ---> Teclas 1 a 9     */
	};

	struct transicao tr;
	static byte estado[ NUMTERM ] = { 0 };
	byte classe; /* classe do evento recebido */
	char codigo;

	/*if (ev.make==2) return;*/
	if (ev.make) {
		switch (ev.tecla) {
			case 0xf: { classe=0; break; } 	/* tecla [F1] */
			case 0xd: { classe=2; break; }	/* tecla [F2] */
			case 0xe: { classe=4; break; }	/* tecla [F3] */
			default: { classe=6; }
		}
	}
	else  {
		switch (ev.tecla) {
			case 0xf: { classe=1; break; } 	/* tecla [F1] */
			case 0xd: { classe=5; break; }	/* tecla [F2] */
			case 0xe:  { classe=3; break; }	/* tecla [F3] */
			default: { classe=7; }
		}
	}
	tr = transicoes[ estado[ term ] ] [ classe ];
	estado[ term ] = tr.ps; /* proximo estado */
	if (tr.ascii >= 1 && tr.ascii <= 4)
		codigo = tascii[ tr.ascii-1 ] [ ev.tecla ];
	else
		codigo = tr.ascii;
	if (codigo!=0) {
		if ( (abuf[term]).f != NULL ) {
			char ret;
			if (ret = proc_accept( codigo,term )) {
				abuf[term].f = NULL;
				abuf[term].ret = ret;
				_signal( &(abuf[term].filled) );
			}
		}
		else {
			teclas[term][tec_ent[term]++] = codigo;
			tec_ent[term]&=0x0f;
			_signal( &skey[term] );
		}
	}
}

static void displayc( term,c )
byte term;
byte c;
{
	disable();
	terminais[term].fila[ terminais[term].e++ ] = c;
	terminais[term].e &= (MAXFILA-1);
	enable();
}

static void displaycmd( term,c )
byte term;
byte c;
{
	disable();
	terminais[term].fila[ terminais[term].e++ ] = c | 0x100;
	terminais[term].e &= (MAXFILA-1);
	enable();
}

void mpc( term,c )
byte term;
byte c;
{
	if (mlin[term]==2) {
		mlin[term]=mcol[term]=0;
		displaycmd( term,1 );
		displayc(term,c);
		return;
	}
	displayc( term,c );
	mcol[term]++;
	if (mcol[term]>=MAXCOLS) {
		if (mlin[term]==1) { mlin[term]=2; mcol[term]=0; }
		else { mlin[term]++; mcol[term]=0; }
	}
}

void mbs( term )
byte term;
{
   if (mcol[term]>0) {
	displaycmd( term,0x10  );
	displayc( term, ' ' );
	displaycmd( term,0x10 );
	mcol[term]--;
   }
}

static void mspos( term,pos )
byte term;
byte *pos;
{
	int cont;
	mcol[term]=*pos++;
	mlin[term]=*pos;
	if (mcol[term]<0 || mcol[term]>16)
		mcol[term]=0;
	if (mlin[term]==0)
		cont=LIN1;	/* endereco inicial no display */
	else
		cont=LIN2;	/* endereco inicial 2 linha	*/
	cont += mcol[term];		/* coluna de 1 a 16 */
	displaycmd(term,cont);
}

int htaccept( buffer,f,echo )
struct fld_desc *f;
char *buffer;
int echo;
{
if (TID==32)
	return (accept_console( buffer,f,echo ));
   {
	struct accept_buf *b;
	int i;
	b = &abuf[TID];
	b->decimal_flg = b->sign = b->digcnt = 0;
	b->echo = echo;
	b->buf = b->buffer = buffer;
	b->cnt = f->len;
	mgetpos(b->pos);
    ALARMES("DDLDD");
	/*
	** echo & 4 --> DARK,
	** echo == 0 --> NO ECHO
	*/
	if (echo & 2) {		/* WITH FILLER */
		for (i=0;i<b->cnt;i++) mputc('.',0);
		if (f->decimals) mputc('.',0);
		msetpos(b->pos);
	}
	if (echo & 8) {		/* WITH UPDATE */
		for (i=0;i<b->cnt;i++)
			mputc(*(buffer+i),0);
		msetpos(b->pos);
	}
	b->f = f;	/* ativa o preenchimento do buffer direto */
	wait( &(b->filled) );
    ALARMES("LDDDD");
	if (b->ret < 0)
		return (b->ret);
	if (f->type=='9') {
		if (!b->decimal_flg) {
			/* alinha campo (inteiro apenas) `a direita */
			if (f->len - b->cnt)
				move_bytes(buffer + b->cnt - f->decimals,
					buffer,b->digcnt);
			/* preenche resto (a esquerda) com zeros */
				set_bytes(buffer,'0',b->cnt - f->decimals);
			/* idem apos o ponto decimal */
			set_bytes(buffer+f->len-f->decimals,
				'0',f->decimals);
		}
		else {
			while (b->cnt) {
				*(b->buf)++ = '0';
				b->cnt--;
				b->digcnt++;
			}
			move_bytes( buffer + f->len - b->digcnt,
				buffer, b->digcnt );
			set_bytes( buffer,'0',f->len - b->digcnt );
		}
	}
	else
		set_bytes( b->buf,' ',b->cnt );
	if (f->type=='9')
		put_sign( f,buffer,b->sign );
	return (int)( b->buf - buffer );
   }
}

char proc_accept( c,term )
char c;
byte term;
{
	char *tmp;
	int i;
	struct accept_buf *b;
	struct fld_desc *f;
	b = &(abuf[term]);
	f = b->f;
	if (c=='-' && b->buf == b->buffer) {
		b->sign=1;
		if (b->echo & 3) mpc(term,'-');
		else if (b->echo & 4) mpc(term,'*');
		return 0;
	}
	if (c==DEL) {
		if (b->buf==b->buffer)
			return -1;
		mspos(term,b->pos);
		tmp=b->buffer;
		while (tmp!=b->buf) {
			if (b->echo & 2) mpc(term,'.');
			else if (b->echo & 8) mpc(term,*tmp);
			else if (b->echo) mpc(term,' ');
			tmp++;
		}	/* limpa campo */
		if (b->decimal_flg) {
			if (b->echo & 2) mpc(term,'.');
			/*else if (b->echo & 8) { } */
					/* nothing to be done */
			else if (b->echo) mpc(term,' ');
		}
		mspos(term,b->pos);
		b->decimal_flg = b->digcnt = 0;
		b->cnt = f->len;
		b->buf = b->buffer;
		return 0;
	}
	if (c==BS) {
		if (b->buf>b->buffer) {
			b->buf--;
			if (b->echo) mbs(term);
			if (b->cnt == f->decimals) {
				b->decimal_flg=0;
				if (b->echo) mbs(term);
			}
			else if (b->digcnt == 1 && b->sign) {
				b->sign=0;
				if (b->echo)
					mbs(term);
			}
			b->cnt++;
			b->digcnt--;
		}
		return 0;
	}
	if (c==CR)
		return 1;
	if ( (f->type=='A') && (isalpha(c) || c==' ') ||
	     (f->type=='X' || f->type=='G')
				&& (isprint(c) || c==' ') ||
	     (f->type=='9') && isdigit(c) ) {
			*(b->buf++) = c;
		if (b->echo & 4) mpc(term,'*');
		else if (b->echo) mpc(term,c);
		b->cnt--;
		if ((b->cnt == f->decimals) && b->cnt) {
			if (b->echo & 4) mpc(term,'*');
			else if (b->echo) mpc(term,',');
			b->decimal_flg++;
		}
		b->digcnt++;
		if (b->cnt)
			return 0;
		else
			return 1;
	}
	if (c==',' && f->type=='9' && !b->decimal_flg) {
		b->decimal_flg++;
		b->cnt = f->decimals;
		if (b->echo & 4) mpc(term,'*');
		else if (b->echo) mpc(term,',');
		if (b->cnt)
			return 0;
		else
			return 1;
	}
}

int accept_console( buffer,f,echo )
struct fld_desc *f;
char *buffer;
int echo;
{
	char *buf,*tmp;
	byte pos[2]; /* para armazenar a posicao apos a msg */
	int cnt,digcnt,i;
	/* cnt 		= digitos que faltam ser digitados,
	** digcnt 	= digitos introduzidos no campo
	*/
	char c,sign=0;
	int decimal_flg=0;
	buf=buffer;
	mgetpos(pos);
	digcnt=0;
	cnt=f->len;
	/*
	** echo & 4 --> DARK,
	** echo == 0 --> NO ECHO
	*/
	if (echo & 2) {		/* WITH FILLER */
		for (i=0;i<cnt;i++) mputc('.',0);
		if (f->decimals) mputc('.',0);
		msetpos(pos);
	}
	if (echo & 8) {		/* WITH UPDATE */
		for (i=0;i<cnt;i++)
			mputc(*(buffer+i),0);
		msetpos(pos);
	}
	while (cnt) {
		c=mgetkey();
		if (c=='-' && buf==buffer) {
			sign=1;
			if (echo & 3) mputc('-',0);
			else if (echo & 4) mputc('*',0);
			continue;
		}
		if (c==DEL) {
			if (buf==buffer) return -1;
			msetpos(pos);
			tmp=buffer;
			while (tmp!=buf) {
				if (echo & 2) mputc('.',0);
				else if (echo & 8) mputc(*tmp,0);
				else if (echo) mputc(' ',0);
				tmp++;
				/* limpa campo */
			}
			if (decimal_flg)
				if (echo & 2) mputc('.',0);
				else if (echo & 8) { } /* nothing to be done */
				else if (echo) mputc(' ',0);
			msetpos(pos);
			decimal_flg=0;
			cnt=f->len;
			digcnt=0;
			buf=buffer;
			continue;
		}
		if (c==BS) {
			if (buf>buffer) {
				buf--;
				if (echo) mputc('\b',0);
				if (cnt==f->decimals) {
					decimal_flg=0;
					if (echo) mputc('\b',0);
				}
				else if (digcnt==1 && sign) {
					sign=0;
					if (echo)
						mputc('\b',0);
				}
				cnt++;
				digcnt--;
			}
			continue;
		}
		if (c==CR) 
			goto final;
		if ( (f->type=='A') && (isalpha(c) || c==' ') ||
		     (f->type=='X' || f->type=='G')
					&& (isprint(c) || c==' ') ||
		     (f->type=='9') && isdigit(c) ) {
				*(buf++) = c;
			if (echo & 4) mputc('*',0);
			else if (echo) mputc(c,0);
			cnt--;
			if (cnt==f->decimals && cnt) {
				if (echo & 4) mputc('*',0);
				else if (echo) mputc(',',0);
				decimal_flg++;
			}
			digcnt++;
			continue;
		}
		if (c==',' && f->type=='9' && !decimal_flg) {
			decimal_flg++;
			cnt=f->decimals;
			if (echo & 4) mputc('*',0);
			else if (echo) mputc(',',0);
			continue;
		}
	}
final:  if (f->type=='9') {
		if (!decimal_flg) {
			/* alinha campo (inteiro apenas) `a direita */
			if (f->len-cnt)
				move_bytes(buffer+cnt-f->decimals,
					buffer,digcnt);
			/* preenche resto (a esquerda) com zeros */
				set_bytes(buffer,'0',cnt-f->decimals);
			/* idem apos o ponto decimal */
			set_bytes(buffer+f->len-f->decimals,
				'0',f->decimals);
		}
		else {
			while (cnt) {
				*buf++ = '0';
				cnt--;
				digcnt++;
			}
			move_bytes(buffer+f->len-digcnt,buffer,digcnt);
			set_bytes(buffer,'0',f->len-digcnt);
		}
	}
	else 
		set_bytes(buf,' ',cnt);
	if (f->type=='9')
		put_sign(f,buffer,sign);
	return (int)(buf-buffer);
}

/* end of MCACC.C */
