/*
**  Cobol Compiler Library -- Accept functions module
**
**  by Rildo Pragana -- 1992
**  P.O. Box 7440 - Recife PE Brazil 50000
**
**  Revision history:
**	29-oct-91 Code implementation
**	24-dec-91 Separate library for files
**	27-apr-92 Separate file for accept functions 
**		(except 'accept from inkey')
**
*/

#include "mcoblib.h"
#define ALARMES(s) liga_alarmes(TID,s)

extern char tascii[4][9];
char teclas[NUMTERM][16];	/* buffer - 16 teclas por terminal */
byte tec_ent[NUMTERM]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	tec_sai[NUMTERM]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

byte sys_buf_char;
SEMAF skey[NUMTERM]={0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,};

/*
**	Processa eventos
**	Armazena ascii correspondente `a sequencia de teclas lida
** 	Pode ser modificada para definir outro layout de teclado.
**
**	NAO DEVE SER USADA COMO SUBROTINA EM PROCESSOS
*/
void process_event( term,ev )
byte term;
struct evento ev;
{
#define ESTADOS 5
#define EVENTOS 8

/*		estados possiveis:
**
**	SHIFT keys = [*], [0], [#]
**
**		1		estado inicial
**		2 x 3	estados de SHIFT		( make e break de SHIFT key )
**	3 x 3	estados de combinacao	( [SHIFT] + [k] :
**											[SHIFT A] & [SHIFT B]
**					eventos possiveis:
**
**	0 - 1,F1 	1 - 0,F1 	2 - 1,F2 	3 - 0,F2
**	4 - 1,F3 	5 - 0,F3 	6 - 1,n 	7 - 0,n
**	|   | |
**	|   | +---> tecla afetada
**	|   +-----> 1= make , 0= break
**	+---------> evento = classe
*/
	static struct transicao {
		byte ps;	/* proximo estado */
		char ascii;	/* codigo a retornar
					0 - nulo
					1 a 4 - tecla de entrada
					outros - ascii */
	} transicoes[ ESTADOS ] [ EVENTOS ] = {

/*							EVENTOS
** ------------------------------------------------------------------------
**	F1	-F1	F2	-F2	F3	-F3	n	-n
** ------------------------------------------------------------------------
*/
	1,0,	0,0,	2,0,	0,0,	3,0,	0,0,	4,1,	0,0,  /* ---> Estado inicial   */
	1,0,	0,0,	2,0,	0,0,	3,0,	0,0,	1,2,	1,0,  /* -+		       */
	1,0,	0,0,	2,0,	0,0,	3,0, 	0,0,	2,3,	2,0,  /*  |-> Estados de SHIFT */
	1,0,	0,0,	2,0,	0,0,	3,0,	0,0,	3,4,	3,0,  /* -+		       */
	1,0,	0,0,	2,0,	0,0,	3,0,	0,0,	0,0,	0,0   /* ---> Teclas 1 a 9     */
	};

	struct transicao tr;
	static byte estado[ NUMTERM ] = { 0 };
	byte classe; /* classe do evento recebido */
	char codigo;

	/*if (ev.make==2) return;*/
	if (ev.make) {
		switch (ev.tecla) {
			case 0xf: { classe=0; break; } 	/* tecla [F1] */
			case 0xd: { classe=2; break; }	/* tecla [F2] */
			case 0xe: { classe=4; break; }	/* tecla [F3] */
			default: { classe=6; }
		}
	}
	else  {
		switch (ev.tecla) {
			case 0xf: { classe=1; break; } 	/* tecla [F1] */
			case 0xd: { classe=5; break; }	/* tecla [F2] */
			case 0xe:  { classe=3; break; }	/* tecla [F3] */
			default: { classe=7; }
		}
	}
	tr = transicoes[ estado[ term ] ] [ classe ];
	estado[ term ] = tr.ps; /* proximo estado */
	if (tr.ascii >= 1 && tr.ascii <= 4)
		codigo = tascii[ tr.ascii-1 ] [ ev.tecla ];
	else
		codigo = tr.ascii;
	if (codigo!=0) {
		teclas[term][tec_ent[term]++] = codigo;
		tec_ent[term]&=0x0f;
		_signal( &skey[term] );
	}
}

int accept( buffer,f,echo )
struct fld_desc *f;
char *buffer;
int echo;
{
	char *buf,*tmp;
	byte pos[2]; /* para armazenar a posicao apos a msg */
	int cnt,digcnt,i;
	/* cnt 		= digitos que faltam ser digitados,
	** digcnt 	= digitos introduzidos no campo
	*/
	char c,sign=0;
	int decimal_flg=0;
	buf=buffer;
	mgetpos(pos);
	if (TID<NUMTERM) ALARMES("LDD");
	digcnt=0;
	cnt=f->len;
	/* 
	** echo & 4 --> DARK, 
	** echo == 0 --> NO ECHO 
	*/
	if (echo & 2) {		/* WITH FILLER */
		for (i=0;i<cnt;i++) mputc('.',0);
		if (f->decimals) mputc('.',0);
		msetpos(pos);
	}
	if (echo & 8) {		/* WITH UPDATE */
		for (i=0;i<cnt;i++)
			mputc(*(buffer+i),0);
		msetpos(pos);
	}
	while (cnt) {
		c=mgetkey();
		if (c=='-' && buf==buffer) {
			sign=1;
			if (echo & 3) mputc('-',0);
			else if (echo & 4) mputc('*',0);
			continue;
		}
		if (c==DEL) {
			if (buf==buffer) return -1;
			msetpos(pos);
			tmp=buffer;
			while (tmp++!=buf) 
				if (echo & 2) mputc('.',0);
				else if (echo & 8) 
					for (i=0;i<f->len;i++)
						mputc(*(buffer+i),0);
				else if (echo) mputc(' ',0);
				/* limpa campo */
			if (decimal_flg)
				if (echo & 2) mputc('.',0);
				else if (echo & 8) { } /* nothing to be done */
				else if (echo) mputc(' ',0);
			msetpos(pos);
			decimal_flg=0;
			cnt=f->len;
			digcnt=0;
			buf=buffer;
			continue;
		}
		if (c==BS) {
			if (buf>buffer) {
				buf--;
				if (echo) mputc('\b',0);
				if (cnt==f->decimals) {
					decimal_flg=0;
					if (echo) mputc('\b',0);
				}
				else if (digcnt==1 && sign) {
					sign=0;
					if (echo)
						mputc('\b',0);
				}
				cnt++;
				digcnt--;
			}
			continue;
		}
		if (c==CR) 
			goto final;
		if ( (f->type=='A') && (isalpha(c) || c==' ') ||
		     (f->type=='X' || f->type=='G')
					&& (isprint(c) || c==' ') ||
		     (f->type=='9') && isdigit(c) ) {
				*(buf++) = c;
			if (echo & 4) mputc('*',0);
			else if (echo) mputc(c,0);
			cnt--;
			if ((cnt==f->decimals) && f->decimals) {
				if (echo & 4) mputc('*',0);
				else if (echo) mputc(',',0);
				decimal_flg++;
			}
			digcnt++;
			continue;
		}
		if (c==',' && f->type=='9' && !decimal_flg) {
			decimal_flg++;
			cnt=f->decimals;
			if (echo & 4) mputc('*',0);
			else if (echo) mputc(',',0);
			continue;
		}
	}
final:  if (f->type=='9') {
		if (!decimal_flg) {
			/* alinha campo (inteiro apenas) `a direita */
			if (f->len-cnt)
				move_bytes(buffer+cnt-f->decimals,
					buffer,digcnt);
			/* preenche resto (a esquerda) com zeros */
				set_bytes(buffer,'0',cnt-f->decimals);
			/* idem apos o ponto decimal */
			set_bytes(buffer+f->len-f->decimals,
				'0',f->decimals);
		}
		else {
			while (cnt) {
				*buf++ = '0';
				cnt--;
				digcnt++;
			}
			move_bytes(buffer+f->len-digcnt,buffer,digcnt);
			set_bytes(buffer,'0',f->len-digcnt);
		}
	}
	else 
		set_bytes(buf,' ',cnt);
	if (f->type=='9')
		put_sign(f,buffer,sign);
	if (TID<NUMTERM) ALARMES("DDL");
	return (int)(buf-buffer);
}

/*
int accept_time( buffer )
char *buffer;
{
	struct tm timep;
	char s[7];
	time(&timep);
	sprintf(s,"%02d%02d%02d",timep.tm_hour,timep.tm_min,timep.tm_sec);
	move_bytes(buffer,s,6);
	return 0;
}

int accept_date( buffer ) 
char *buffer;
{
	struct tm timep;
	char s[7];
	time(&timep);
	sprintf(s,"%02d%02d%02d",(timep.tm_year)%100,
		timep.tm_mon,timep.tm_mday);
	move_bytes(buffer,s,6);
	return 0;
}
*/
/* end of MCACCEPT.C */
