/*
**  Cobol Compiler Library -- General Utility Module
**
**  by Rildo Pragana -- 1991
**  P.O. Box 7440 - Recife PE Brazil 50000
**
**  Revision history:
**	29-oct-91 Code implementation
**	24-dec-91 Separate library for files
**
*/

#include <math.h>
#include "mcoblib.h"

extern PROCESS *running;
extern SEMAF sdsp[];
#define min(x,y) ((x)<(y) ? (x) : (y))

int lib_debug = 0; /* =1 for debugging  */
struct task_table {
	void (*function)();
	int stack_words;
	char name[9];
	char terminal;
}*init_tbl=NULL;
struct task_table *load_task_table( );
char current_dir[65]="";

extern byte tec_ent[NUMTERM];
extern byte tec_sai[NUMTERM];
extern SEMAF skey[NUMTERM];
extern byte teclas[NUMTERM][16];

byte sys_buf_char = 0;
SEMAF syskey = {0,NULL};

void liga_alarmes ( alarmes )
char *alarmes;
{
 int term=TID;
 unsigned int p=1;
 int t;
 if (term>=32) return;
 for ( t=0 ; t<=5 ; t++ ) {
	  if ( alarmes[t] == 'L')
		 terminais[term].alarme = terminais[term].alarme | p;
	   else
	     if ( alarmes[t] == 'D' )
		terminais[term].alarme = terminais[term].alarme & ~p;
	   else
		break;
	 p=p<<1;
	 }
	 terminais[term].alarme |= 0x80; /* indica alarme pendente */
}

/*
**	rotina chamada pela Int2f para bufferizar uma tecla da console
*/
void put_sysch( c )
byte c;
{
	sys_buf_char = c;
	_signal(&syskey);
}

/*
**	retorna uma tecla ja processada
*/
byte mgetkey();

byte htgetc() { return mgetkey(); }

byte mgetkey( )
{
	char c;
    unsigned char tid=TID;
    if (tid==32) {
		wait(&syskey);
		return sys_buf_char;
	}
    if (tid>32) return 0;
    if (tec_ent[tid]==tec_sai[tid])
        wait( &skey[tid] );
	disable();
    c=teclas[tid][tec_sai[tid]++];
    tec_sai[tid]&=0x0f;
	enable();
	return c;
}

/*
**  Envia um caracter ao display do terminal
*/
static void displaybyte( term, c )
byte term,c;
{
	disable();
	if (((terminais[term].s-terminais[term].e)==1)
		||(terminais[term].s==0 && terminais[term].e==MAXFILA))
			wait (&sdsp[term]);
	terminais[term].fila[ terminais[term].e++ ] = c;
	terminais[term].e &= (MAXFILA-1);
	enable();
}

static void displaycomm( term, c )
byte term,c;
{
	disable();
	if (((terminais[term].s-terminais[term].e)==1)
		||(terminais[term].s==0 && terminais[term].e==MAXFILA))
			wait (&sdsp[term]);
	terminais[term].fila[ terminais[term].e++ ] = c | 0x100;
	terminais[term].e &= (MAXFILA-1);
	enable();
}

/* constantes relacionadas com a programacao do display */
#define CLRSCR	1
#define LIN1	0x80
#define LIN2	0xc0
#define CURLEFT 04
#define CURRIG	06
#define CUROFF  0x0c
#define CURON   0x0e
#define CURDL   0x10
#define MAXCOLS 16

/*
**	limpa a tela do mt100
*/
static void mclrscr( )
{
	byte term=TID;
	if ((terminais[term].on_line == 0) ||
		(terminais[term].on_line && 2)) {
		/* envia sequencia de inicializacao do display */
		displaycomm(term,0x38);
		displaycomm(term,0x38);
		displaycomm(term,0x05);
		displaycomm(term,0x0e);
		displaycomm(term,0x13);
		displaycomm(term,0x01);
		displaycomm(term,0x06);
		displaycomm(term,0x80);
		terminais[term].on_line=1;
	}
	else
		displaycomm( term,1 );
}

/*
**	envia sequencia para "backspace" no display
*/
static void mcurback( )
{
	byte term=TID;
	/* DESLOCA PARA ESQUERDA */
	displaycomm( term,0x10 );
	/* LIMPA CARACTER  */
	displaybyte( term, ' ' );
	/* DESLOCA PARA ESQUERDA */
	displaycomm( term,0x10 );
}

/*
**	posiciona o cursor em dada posicao
*/
static void mgotolc( lin, col )
byte lin,col;
{
	byte term=TID;
	int cont;
	if (!(col>0 && col<17) )
		col=1;
	if (lin==1)
		cont=LIN1;	/* endereco inicial no display */
	else
		cont=LIN2;	/* endereco inicial 2 linha	*/
	cont+=(col-1);		/* coluna de 1 a 16 */
	displaycomm( term,cont );
}

byte mcol[NUMTERM];
byte mlin[NUMTERM];
/*
** armazena posicao corrente
*/
void mgetpos( pos )
byte *pos;
{
	byte term=TID;
	if (term>=32) {
		pos[0]=(byte)wherex(); pos[1]=(byte)wherey(); return;
	}
	*pos++=mcol[term];
	*pos=mlin[term];
}

/*
** reposiciona o cursor
*/
void msetpos( pos )
byte *pos;
{
	byte term=TID;
	if (term>=32) {
		gotoxy(pos[0],pos[1]); return;
	}
	mcol[term]=*pos++;
	mlin[term]=*pos;
	mgotolc((*pos)+1,*(pos-1)+1);
}

/*
**	envia um caracter para o display, inclusive especiais
*/
void mputc( byte, byte );

void htputc( c,wait )
byte c,wait; { mputc(c,wait); }

void mputc( c,wait )
byte c,wait;
{
	byte term=TID;
	if (term>=32) {
		if (c=='\f') clrscr();
		else
			putchar(c);
		return;
	}
	if (c=='\f') {
		mclrscr();
		mlin[term]=mcol[term]=0;
	}
	else if (mlin[term]==2) {
		if (wait) {
			ALARMES("DLD");
			mgetkey();
		}
		mlin[term]=mcol[term]=0;
		mclrscr();
		mputc(c,wait);
	}
	else if ( c & 0x80 ){ 	/* checa se d7=1	*/
		displaybyte( term,c );
		mcol[term]++;
		if (mcol[term]>=MAXCOLS)
			mputc( '\n',wait );
	}
	else if (c>=' ') {
		displaybyte( term,c);
		mcol[term]++;
		if (mcol[term]>=MAXCOLS)
			mputc( '\n',wait );
	}
	else
		switch (c) {
		case '\n': /* avanca uma linha */
			if (mlin[term]==1) {
				/* linha estourada, aguarde tecla */
				mlin[term]=2; mcol[term]=0;
			}
			else
			{
				mlin[term]++; mcol[term]=0; mgotolc(2,1);
			}
			break;
		case '\r': /* retorna para o inicio desta linha */
			mgotolc( mlin[term]+1, 1 );
			mcol[term]=0;
			break;
		case '\b': /* backspace */
			if (mcol[term]>0) {
				mcurback();
				mcol[term]--;
			}
			break;
		default:
			displaybyte( term,c );
			mcol[term]++;
			if (mcol[term]>=MAXCOLS)
				mputc( '\n',wait );
			break;
		}
}

/*
**	envia uma sequencia de caracteres para o display
*/
void mputstr( s )
char *s;
{
	if (!*s)
		return;
	while( *s ){
		mputc( *s++,1 );
	}
}

void goxy( col,lin )
int col,lin;
{
	byte term=TID;
	if (term>=32) {
		gotoxy(col,lin);
		return;
	}
	mcol[term]=(byte)(col-1);
	mlin[term]=(byte)(lin-1);
	mgotolc(lin,col);
}

PROCESS task[68];    /* array de processos para os terminais */

static void inicia_processos( )
{
	int i;
	init_tbl=load_task_table();
	for (i=0;i<NUMTERM+5;i++)
		if ((init_tbl+i)->stack_words)
			start_task(
				task+i,
				(init_tbl+i)->stack_words,
				(init_tbl+i)->function,
				(init_tbl+i)->name,
				(init_tbl+i)->terminal
			);
/*    init_spooler();*/
}

/********* variaveis para TESTE **************/
/*unsigned int tstack[1040];*/
/*char ch[2];
unsigned int *tstack;
void create_task( unsigned int,void *,void *,unsigned int,char);

void tfunction() {
	mgetkey();
	mputc('\f',1);
	while(1) 
		mputc(mgetkey(),1);
}*/

void main( argc,argv )
int argc;
char **argv;
{
/*	unsigned int *dataseg;
	dataseg = ch;*/
/*	int fp;
	void *p;
*/
	mtnet_init(argc, argv );	/* inicializacao */
	_getcwd(current_dir,64);
	strcat(current_dir,"\\");
/*printf("\nDiretorio corrente: %s\n",current_dir); getch();
*/
	inicia_processos();
/*	tstack = (unsigned int *)malloc(2000);
	create_task(1000,tstack,tfunction,FP_SEG(dataseg),0);
*/
printf("\nPSP corrente: %04x\n",_psp);
/*	fp=_creat("mttask.dat",0);
	p=(void *)init_tbl;
	_write(fp,(char *)&p,4);
	p=(void *)task;
	_write(fp,(char *)&p,4);
	_close(fp);
*/
	mtnet_install();
}

void stop_run( ) {
	ALARMES("DDD");
	mprintf("\f*** HandTerm ***\n    OFFLINE");
	while(1) mgetkey();
}

void accept_inkey( buffer ) 
char *buffer;
{
	char c,s[4];
    unsigned char tid=TID;
    if (tec_ent[tid]==tec_sai[tid])
		c=0;
	else {
		disable();
        c=teclas[tid][tec_sai[tid]++];
        tec_sai[tid]&=0x0f;
		enable();
	}
	sprintf(s,"%03.3d",c);
	move_bytes(buffer,s,3);
}

void runtime_err( ) { }

void init_terminal( ) {
    unsigned char tid=TID;
    if (tid>32) return;
    if (tid==32) mgetkey();
    if (tid<NUMTERM) {
        ALARMES("LLLLL");
        mprintf("\f\xdb\xdb\xdb HandTerm \xdb\xdb\xdbTerminal ativo!");
    }
        while(mgetkey()!=DEL);
	mputc('\f',0);
}

void display( buffer,len )
char *buffer;
int len;
{ /* anteriormente disp_glob() */
	while (len--)
		mputc(*buffer++,1);
}

void newline( ) {
	if (TID<=32)
		mputc('\n',1);
}

void display_erase( ) {
	mputc('\f',1);
}

extern unsigned char prtfila[32][64];
extern SEMAF sprt[];

void pputc( n,c )
byte c;
byte n;
{
    byte term;
    term = -n-64;
    disable();
    if (((terminais[term].ps-terminais[term].pe)==1)
        ||(terminais[term].ps==0 && terminais[term].pe==63))
            wait (&sprt[term]);
    prtfila[term][terminais[term].pe++] = c;
    terminais[term].pe &= 63;
    enable();
}

void pnewline( n )
int n; {
    pputc(n,0x0d);
    pputc(n,0x0a);
}

/* end of HTLIBDIR.C */
