/*
**  Cobol Compiler Library -- File Handling Module
**
**  by Rildo Pragana -- 1991
**  P.O. Box 7440 - Recife PE Brazil 50000
**
**  Revision history:
**	29-oct-91 Code implementation
**	24-dec-91 Separate library for files
**
*/

#include "mcoblib.h"
#include "math.h"

#define min(x,y) ((x)<(y) ? (x) : (y))
#define PRN_HANDLE *((int *)&(v->filepos))

int spool_number=0;
extern char current_dir[];

void open_prn( prn_port,f,v )
int prn_port;
struct file_desc *f;
struct var_fd *v;
{
	char fname[65],tmp[20];
	v->splrec = spool_number++;
	strcpy(fname,current_dir);
	sprintf(tmp,"SPL%05d\.\$MT",v->splrec);
	strcat(fname,tmp);
	PRN_HANDLE = _creat(fname,0);
	f->organization=4; /* special for spooler */
	f->users=prn_port; /* where to save prn_port */
}

void close_prn( f,v ) 
struct file_desc *f;
struct var_fd *v;
{
	char fname[65],tmp[20];
	_close(PRN_HANDLE);
	sprintf(fname,"%%%c",f->users+'0');
	strcat(fname,current_dir);
	sprintf(tmp,"SPL%05d\.\$MT",v->splrec);
	strcat(fname,tmp);
	submit_print(fname);
}

int write_prt( f,v )
struct file_desc *f;
struct var_fd *v;
{
	if (_write(PRN_HANDLE,RECORD,f->reclen)==-1) return 30;
	if (_write(PRN_HANDLE,"\x0d\x0a",2)==-1) return 30;
	return 0;
}

int save_status( status, rt )
int rt;
char *status;
{
	*status++ = (rt/10)+'0';
	*status = (rt%10)+'0';
	return rt;
}

static int sortnumber=1;
static char *sortdesc=NULL;

int keycmp( s1,s2 )
char *s1;
char *s2;
{
	if (sortdesc==NULL)
		return strcmp( s1,s2 );
	else {
		unsigned char i;
		unsigned char direction;
		char *sd;
		sd=sortdesc;
		while (direction=*sd++) {
			for (i=0;i<*sd;i++) {
				if (*s1 < *s2) {
					if (direction==1) return -1;
					else return 1;
				}
				if (*s1++ > *s2++) {
					if (direction==1) return 1;
					else return -1;
				}
			}
			sd++;
		}
		return 0;
	}
}

int sort_open( f,v )
struct file_desc *f;
struct var_fd *v;
{
	char fname[64];
	char snum[6];
	sprintf(snum,"%05d",sortnumber++);
	strcpy( fname,current_dir );
	strcat( fname,"srt" );
	strcat( fname,snum );
	strcat( fname,".$$d" );
	v->filepos=0;
	v->releasing=1;
	if ((f->handle)>0) return 30;	/* um de cada vez */
	if ((f->handle = _creat( fname,0))==-1) return 30;
	strcpy( fname,current_dir );
	strcat( fname,"srt" );
	strcat( fname,snum );
	strcat( fname,".$$i" );
	make_index(fname,&(f->index),1);
}

int sort_return( f,v )
struct file_desc *f;
struct var_fd *v;
{
	int rt;
	if (v->releasing) {
		if ( first_key( &(f->index) )!=IX_OK )
			return 23;
		next_key( &(v->e),&(f->index) );
		v->releasing=0;
		v->filepos=v->e.recptr;
	}
	if (v->filepos<0) {
		_close(f->handle);
		f->handle=0;
		close_index(&(f->index));
		return 10; /* end of file */
	}
	if (_lseek(f->handle,v->e.recptr,SEEK_SET)==-1) return 30;
	rt=_read( f->handle, RECORD, f->reclen );
	if (rt==-1) return 30; /* read error */
	if (next_key( &(v->e),&(f->index))!=IX_OK)
		v->filepos=-1;
	else
		v->filepos=v->e.recptr;
	return 0;
}

int sort_release( f,v,sd,... )
struct file_desc *f;
struct var_fd *v;
char *sd;
{
	va_list argp;
	char *fld;
	char *key;
	if (_write(f->handle,RECORD,f->reclen)==-1) return 30;
	sortdesc = sd;
	key = v->e.key;
	va_start( argp,sd );
	while(*sd++) {
		fld = va_arg( argp,char * );
/*mprintf("\fKey: %4.4s,%d",fld,*sd); mgetkey();*/
		move_bytes( key,fld,*sd );
		key += *sd++;
	}
	*key=0;
	v->e.recptr = v->filepos;
	v->filepos += f->reclen;
	if (add_key( &(v->e), &(f->index) )!=IX_OK)
		{ sortdesc=NULL; return 22; }
	sortdesc=NULL;
	return 0;
}

int mc_open( f,v,mode )
struct file_desc *f;
struct var_fd *v;
int mode;
{
	char fname[64];
	unsigned long i,seekpos,seekold;
	int ac,ix;
#if	MICROSOFT == 2
	if (f->organization==1)
		v->filepos=64;
	else
		v->filepos=0;
#else
	v->filepos=0; 		/* set record counter to the first one */
#endif
	if (f->handle<=0) { 	/* do not open if it's already opened
				   and test error (-1) handle number */
		if (strchr(f->filename,'\\')==NULL)
			strcpy( fname,current_dir );
		else
			strcpy( fname, "" );
		strcat( fname,f->filename );
		strcat( fname,".arq" );
		switch(mode) {
		case 1: /* open input */
			if ((f->handle = _open( fname,
				O_RDONLY|O_BINARY))==-1) return 30;
				break;
		case 2: /* open i/o */
			if ((f->handle = _open( fname,
				O_RDWR|O_BINARY ))==-1) return 30;
				break;
		case 3: /* open output */
			if ((f->handle = _creat( fname,0))==-1) return 30;
				break;
		case 4: /* open extend */
			if ((f->handle = _open( fname,
				O_WRONLY|O_BINARY|O_APPEND ))==-1)
					return 30;
				break;
		default: return 97; /* unknown mode */
		}
		f->users=1; 	/* first user for this file */
	}
	else  {
		f->users++;
		return 0;
	}
	if (f->organization==1) {
		if (strchr(f->filename,'\\')==NULL)
			strcpy( fname,current_dir );
		else
			strcpy( fname, "" );
		strcat( fname,f->filename );
		strcat( fname,".ixd" );
		ac=_access(fname,0);
		if (ac==0 && mode!=3) {
			ix=open_index(fname,&(f->index),0);
			next_key( &(v->e),&(f->index) );
		}
		if (ac!=0 || ix!=IX_OK || mode==3) {
			/* try to re-index file */
		    make_index(fname,&(f->index),1);
		    if (mode!=3) {
#if	MICROSOFT==2
		if (f->organization==1)
			seekpos=64;
		else
			seekpos=0;
#else
			seekpos=0;
#endif
			for (i=0;;i++) {
				_lseek( f->handle, seekpos, SEEK_SET );
				if ( _read( f->handle, RECORD,
					f->reclen ) != f->reclen ) {
					chsize( f->handle, seekpos );
					break;
				}
	if (*(RECORD) == 0) {
		seekold=seekpos;
		while (*(RECORD)==0) {
#if MICROSOFT == 2
			if (f->organization==1)
				seekpos += f->reclen+3;
			else
				seekpos += f->reclen;
#else
			seekpos += f->reclen;
#endif
			_lseek( f->handle, seekpos, SEEK_SET );
			if ( _read( f->handle, RECORD,
				f->reclen ) != f->reclen ) {
				chsize( f->handle, seekold );
				goto endreg;
			}
		}
		_lseek( f->handle, seekold, SEEK_SET );
		_write( f->handle, RECORD,f->reclen );
		_lseek( f->handle, seekpos, SEEK_SET );
		_write( f->handle, "\0", 1 );
		seekpos=seekold;
}
				move_bytes( v->e.key,RECORD+f->rec_index,
					 f->ixd_desc->len );
				*(v->e.key + f->ixd_desc->len)=0;
				v->e.recptr = seekpos;
				if (add_key( &(v->e), &(f->index) ) != IX_OK)
				{
					mprintf("\fErro na adicao  de chave");
					mgetkey();
					return 30;
				}
#if MICROSOFT==2
				if (f->organization==1)
					seekpos += f->reclen+3;
				else
					seekpos += f->reclen;
#else
				seekpos += f->reclen;
#endif
			} /* end of re-indexation */
		    }
endreg:		    first_key( &(f->index) );
		    next_key( &(v->e),&(f->index) );
		}
	}
/*	if (f->organization==4 && mode!=3)
		_lseek(f->handle,-1L,SEEK_END); */
		/* overwrite 0x1a at end */
	return 0;
}

void mc_commit( handle )
int handle;
{
	union REGS r;
	r.x.ax=0x6800;
	r.x.bx=handle;
	int86(0x21,&r,&r);
}

int mc_close( f
#ifdef MICROBASE
	,v	/* required for MICROBASE format */
#endif
	)
struct file_desc *f;
#ifdef MICROBASE
struct var_fd *v;
#endif
{
#ifdef MICROBASE
	long fill;
	char *buf;
	int i;
#endif
	if (!(--(f->users))) {		/* last user, please close it */
#ifdef MICROBASE
		fill = 1024 - _filelength(f->handle) % 1024;
		buf=RECORD;
		for (i=0;i<f->reclen;i++)
			 *buf++ = -i;
		while (fill) {
			if (fill>f->reclen) {
				_write(f->handle,RECORD,f->reclen);
				fill -= f->reclen;
			}
			else {
				_write(f->handle,RECORD,(unsigned)fill);
				fill = 0;
			}
		}
#endif
/*		if (f->organization==4)
			_write(f->handle,"\x1a",1);*/
		_close(f->handle);
		f->handle=0;
		if (f->organization==1)
			close_index(&(f->index));
	}
	else { /* commit data file */
/*		if (f->organization==4)
			_write(f->handle,"\x1a",1);*/
		mc_commit(f->handle);
	}
	return 0;
}

int mc_read( f,v,... )
struct file_desc *f;
struct var_fd *v;
{
	int rt;
	va_list args;
	if (f->organization==3 && f->access_mode!=1) {
		/* relative, not sequential */
		unsigned recn;
		long fpos;
		va_start(args,v);	/* inicia 'args' apos 'v' */
		recn=va_arg(args,unsigned);
		fpos= (long)(recn-1) * (long)f->reclen;
		rt=(int)_lseek(f->handle,fpos,SEEK_SET);
		if (rt==-1) return 31; /* seek error */
		rt = _read( f->handle, RECORD, f->reclen );
		if (rt==0) return 10; /* end of file */
		if (rt==-1) return 30; /* read error */
		v->filepos=fpos;
		va_end(args);
		return 0;
	}
	if (f->organization!=1 && f->access_mode==1)
	{ /* not indexed, sequential */
next_rec:	rt=(int)_lseek(f->handle,v->filepos,SEEK_SET);
		if (rt==-1) return 31; /* seek error */
		rt = _read( f->handle, RECORD, f->reclen );
		if (rt==0) return 10; /* end of file */
		if (rt==-1) return 30; /* read error */
#if MICROSOFT==2
		if (f->organization==1)
			v->filepos += f->reclen+3;
		else
			v->filepos += f->reclen;
#else
		v->filepos += f->reclen;
#endif
		if (*(RECORD) == 0) goto next_rec; /* deleted record */
	}
	else if (f->organization==1 && f->access_mode!=1)
	{ /* indexed, not sequential */
		move_bytes( v->e.key,RECORD+f->rec_index,
			f->ixd_desc->len );
		*(v->e.key + f->ixd_desc->len)=0;
/*mprintf("\fv->e.key %04x:%04x -> '%s'",FP_SEG(v->e.key[0]),
	FP_OFF(v->e.key[0]),v->e.key); mgetkey();*/
		if (find_key( &(v->e), &(f->index))!=IX_OK)
			return 23; /* record not found */
		if(_lseek(f->handle,v->e.recptr,SEEK_SET)==-1)
			return  31;
		v->filepos=v->e.recptr;
		rt=_read( f->handle, RECORD, f->reclen );
		if (rt==0) return 10; /* end of file */
		if (rt==-1) return 30; /* read error */
	}
	else if (f->organization==1 && f->access_mode==1)
	{ /* indexed, sequential */
		if (find_key( &(v->e),&(f->index))!=IX_OK) return 23;
		if (next_key( &(v->e),&(f->index))!=IX_OK) return 10;
		if (_lseek(f->handle,v->e.recptr,SEEK_SET)==-1) return 30;
		v->filepos=v->e.recptr;
		rt=_read( f->handle, RECORD, f->reclen );
		if (rt==0) return 10; /* end of file */
		if (rt==-1) return 30; /* read error */
	}
	return 0;
}

int mc_read_next( f,v )
struct file_desc *f;
struct var_fd *v;
{
	int rt;
	if (f->organization==3) {
		/* relative */
		rt=(int)_lseek(f->handle,v->filepos,SEEK_SET);
		if (rt==-1) return 31; /* seek error */
		rt = _read( f->handle, RECORD, f->reclen );
		if (rt==0) return 10; /* end of file */
		if (rt==-1) return 30; /* read error */
		return 0;
	}
	if (v->filepos==-1) return 10;
	if (_lseek(f->handle,v->filepos,SEEK_SET)==-1) return 30;
	rt=_read( f->handle, RECORD, f->reclen );
	if (rt==0) return 32; /* end of file */
	if (rt==-1) return 30; /* read error */
	if (next_key( &(v->e),&(f->index))!=IX_OK)
			v->filepos = -1;
	else
		v->filepos = v->e.recptr;
	return 0;
}

int mc_write( f,v )
struct file_desc *f;
struct var_fd *v;
{
	long seekpos;
	if (f->organization != 4) {
		/* seekpos = _filelength(f->handle); */
		if (_lseek(f->handle,0,SEEK_END)==-1L) return 21;
		seekpos=_tell(f->handle);
	}
	else {
		if (_lseek(f->handle,-1L,SEEK_END)==-1L)
			if (_lseek(f->handle,0,SEEK_END)==-1L) return 21;
	}
	if (_write(f->handle,RECORD,f->reclen)==-1) return 30;
	if (f->organization==4) /* printer or line sequential */
		_write(f->handle,"\x0d\x0a\x1a",3);
	if (f->organization==1) {
		move_bytes( v->e.key,RECORD+f->rec_index,
				 f->ixd_desc->len );
		*(v->e.key + f->ixd_desc->len)=0;
			/* null at end of string */
		v->e.recptr = seekpos;
		if (add_key( &(v->e), &(f->index) )!=IX_OK) return 22;
	}
	return 0;
}

int write_adv( f,v,opt,... )
struct file_desc *f;
struct var_fd *v;
int opt;
{
	va_list argp;
	int retcode;
	struct fld_desc *cnt_desc;
	char *cnt_buf;
	int lines=0;
	va_start( argp, opt );
	if ( opt>0 ) {
		cnt_desc = va_arg( argp,struct fld_desc * );
		cnt_buf = va_arg( argp,char * );
		lines=get_index( cnt_desc,cnt_buf );
		if (opt==1) {
			retcode=write_prt( f,v );
			for (;lines;lines--)
				_write(PRN_HANDLE,"\x0d\x0a",2);
		}
		else { /* opt==2 */
			for (;lines;lines--)
				_write(PRN_HANDLE,"\x0d\x0a",2);
			retcode=write_prt( f,v );
		}
	}
	else if ( opt==-1 ) { /* before advancing page */
		retcode=write_prt( f,v );
		_write(PRN_HANDLE,"\x0c",1);
	}
	else if ( opt==-2 ) { /* after advancing page */
		_write(PRN_HANDLE,"\x0c",1);
		retcode=write_prt( f,v );
	}
	else /* only normal write (error?) */
		retcode=write_prt( f,v );
	return retcode;
}

int mc_rewrite( f,v ) 
struct file_desc *f;
struct var_fd *v;
{
	if (f->organization != 1) { /* not indexed */
		if (f->organization != 3) /* not relative */
			v->filepos -= f->reclen;
		if (_lseek(f->handle,v->filepos,SEEK_SET)==-1L) return 21;
		if (_write(f->handle,RECORD,f->reclen)==-1) return 30;
	}
	if (f->organization==1) {
		/* indexed */
		if (find_key( &(v->e),&(f->index))!=IX_OK) return 23;
		if (memcmp( v->e.key,RECORD+f->rec_index,
				 f->ixd_desc->len )) return 99;
		if (_lseek(f->handle,v->e.recptr,SEEK_SET)==-1L) return 21;
		if (_write(f->handle,RECORD,f->reclen)==-1) return 30;
	}
	return 0;
}

int mc_start( f,v ) 
struct file_desc *f;
struct var_fd *v;
{
	if (f->organization==1) {
		move_bytes( v->e.key,RECORD+f->rec_index,
			f->ixd_desc->len );
		*(v->e.key + f->ixd_desc->len)=0;
		if (locate_key( &(v->e),&(f->index))==EOIX) {
			v->filepos = -1;
			return 23;
		}
		else
			v->filepos = v->e.recptr;
/*		if (_lseek(f->handle,v->e.recptr,SEEK_SET)==-1L) return 21;
		if (_read( f->handle, RECORD, f->reclen )==-1) return 30;
*/	}
	return 0;
}

int mc_delete( f,v )
struct file_desc *f;
struct var_fd *v;
{
	if (f->organization==1) { /* indexed */
		if (find_key( &(v->e),&(f->index))!=IX_OK) return 23;
		if (_lseek(f->handle,v->e.recptr,SEEK_SET)==-1L) return 21;
		*((char *)RECORD) = 0; /* set to null first char */
		if (_write(f->handle,RECORD,f->reclen)==-1) return 30;
		delete_key( &(v->e), &(f->index) );
	}
	else {
		if (f->organization != 3) /* not relative */
			v->filepos -= f->reclen;
		if (v->filepos < 0) return 29;
		if (_lseek(f->handle,v->filepos,SEEK_SET)==-1L) return 21;
		*((char *)RECORD) = 0;
		if (_write(f->handle,RECORD,f->reclen)==-1) return 30;
	}
	return 0;
}

