/*
**  Cobol Compiler Library -- General Utility Module
**
**  by Rildo Pragana -- 1991
**  P.O. Box 7440 - Recife PE Brazil 50000
**
**  Revision history:
**	29-oct-91 Code implementation
**	24-dec-91 Separate library for files
**
*/

#define SIMULA 1   /* 0=normal, 1=simulador */
#include <math.h>
#include "mcoblib.h"

#define min(x,y) ((x)<(y) ? (x) : (y))

int lib_debug = 0; /* =1 for debugging  */
struct task_table {
	void (*function)();
	int stack_words;
	char name[9];
	char terminal;
}*init_tbl=NULL;
struct task_table *load_task_table( );
void submit_print( /*char *fname*/ );
char current_dir[65]="";

#if SIMULA == 1
/******************* rotinas usadas na simulacao do MT100 *****************/
#define TID 0
#define NUMTERM 20
typedef unsigned char byte;

#define DEL 0x7f
#define CR '\r'
#define BS '\b'
#define ALARMES(s) liga_alarmes(TID,s)

void mputc( /*byte c, byte wait*/ );

void submit_print( /*char *s*/ ) { }

void main() {
	int term;
	_getcwd(current_dir,64);
	strcat(current_dir,"\\");
	clrscr();
	init_tbl=load_task_table();
	printf("Simulador de Microterminais MT100\n");
	printf("Entre endereco do terminal a emular: ");
	scanf("%d",&term);
	if (((init_tbl+term)->function)==NULL)	{
		printf("\nEndereco nao inicializado\n");
		exit(1);
	}
	(*((init_tbl+term)->function))();
}

void liga_alarmes( dummy,s ) 
unsigned char dummy; /* dummy TID for compatibility */
char *s;
{
	int x,y;
	x=wherex();
	y=wherey();
	gotoxy(25,1);
	while(*s) {
		if (*s=='L') putchar(178);
		else putchar(176);
		putchar(' ');
		s++;
	}
	gotoxy(x,y);
}

void mputc( c, wait )
byte c,wait;
{
	if (c=='\f') clrscr();
	else if (wherey()==3) {
		if (wait) {
			ALARMES("DLD");
			mgetkey();
			clrscr();
		}
		else goxy(1,1); /* modificado p/ficar mais simples */
		mputc(c,wait);
	}
	else if (c>=' ') {
		putchar(c);
		if (wherex()>=17)
			mputc( '\n',wait );
	}
	else
		putchar(c);
}
void mputstr( s ) 
char *s;
{ while(*s) mputc( *s++,1 ); }

/*void mprintf( fmt, ...)
char *fmt;
{
	va_list argptr;
	char str[129];
	va_start( argptr, fmt );
	vsprintf( str, fmt, argptr );
	mputstr( str );
	va_end( argptr );
}*/
void mcurback( ) {
	putchar('\b');
}
byte mgetkey( ) {
  byte c;
  c=(byte)getch();
  if (c==0x18) exit(0);
  return c;
}
void mgetpos( pos ) byte *pos; { pos[0]=wherex(); pos[1]=wherey(); }
void msetpos( pos ) byte *pos; { gotoxy(pos[0],pos[1]); }
void goxy( col,lin ) int col,lin; { gotoxy( col, lin ); }
void stop_run( ) {
	ALARMES("DDD");
	mprintf("\f* TERMINAL OFF *");
	while(mgetkey() != 0x18);
	exit(0);
}

void accept_inkey( buffer )
char *buffer;
{
	char c,s[4];
	if (kbhit()==0)
		c=0;
	else {
		c=getch();
		if (c==0)
			if ((c=getch())=='S') c=DEL;
	}
	sprintf(s,"%03.3d",c);
	move_bytes(buffer,s,3);
}

/**************************************************************************/

#elif SIMULA == 0

/********* Rotinas para uso On-line com Microterminais ********************/

#define ALARMES(s) liga_alarmes(TID,s)

extern char tascii[4][9];
static char teclas[NUMTERM][16];	/* buffer - 16 teclas por terminal */
byte tec_ent[NUMTERM]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	tec_sai[NUMTERM]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

SEMAF syskey={0,NULL};
byte sys_buf_char;
SEMAF skey[NUMTERM]={0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,
		     0,NULL,0,NULL,0,NULL,0,NULL,};

/*
**	Processa eventos
**	Armazena ascii correspondente `a sequencia de teclas lida
** 	Pode ser modificada para definir outro layout de teclado.
**
**	NAO DEVE SER USADA COMO SUBROTINA EM PROCESSOS
*/
void process_event( term,ev )
byte term;
struct evento ev;
{
#define ESTADOS 17
#define EVENTOS 8

/*		estados possiveis:
**
**	SHIFT keys = [*], [0], [#]
**
**		1		estado inicial
**		2 x 3	estados de SHIFT		( make e break de SHIFT key )
**	3 x 3	estados de combinacao	( [SHIFT] + [k] :
**											[SHIFT A] & [SHIFT B]
**											[SHIFT A] & [SHIFT C]
**											[SHIFT A] & [n] )
**	1		estado para teclas numericas	( n )
**
**					eventos possiveis:
**
**	0 - 1,* 	1 - 0,* 	2 - 1,0 	3 - 0,0
**	4 - 1,# 	5 - 0,# 	6 - 1,n 	7 - 0,n
**	|   | |
**	|   | +---> tecla afetada
**	|   +-----> 1= make , 0= break
**	+---------> evento = classe
*/

	static struct transicao {
		byte ps;	/* proximo estado */
		char ascii;	/* codigo a retornar
					0 - nulo
					1 a 4 - tecla de entrada
					outros - ascii */
	} transicoes[ ESTADOS ] [ EVENTOS ] = {

/*							EVENTOS
** ------------------------------------------------------------------------
**	*		-*		0		-0		#		-#		n		-n
** ------------------------------------------------------------------------
*/
	1,0,	0,0,	2,0,	0,0,	3,0,	0,0,	4,1,	0,0,  /* ---> Estado inicial   */
	0,0,	0,DEL,	7,'/',	0,0,	6,'?',	0,0,	5,2,	0,0,  /* -		       */
	8,'.',	0,0,	0,0,	0,'0',	10,',', 0,0,	9,3,	0,0,  /*  |-> Estados de SHIFT */
	11,'-', 0,0,	13,BS,	0,0,	0,0,	0,CR,	12,4,	0,0,  /* -		       */
	0,0,	0,0,	0,0,	0,0,	0,0,	0,0,	0,0,	0,0,  /* ---> Teclas 1 a 9     */
	0,0,	0,0,	0,0,	0,0,	0,0,	0,0,	0,0,	14,0, /* -		       */
	0,0,	0,0,	0,0,	0,0,	0,0,	14,0,	0,0,	0,0,  /*  |		       */
	0,0,	0,0,	0,0,	14,0,	0,0,	0,0,	0,0,	0,0,  /*  |		       */
	0,0,	15,0,	0,0,	0,0,	0,0,	0,0,	0,0,	0,0,  /*  |		       */
	0,0,	0,0,	0,0,	0,0,	0,0,	0,0,	0,0,	15,0, /*  |-> Estados de       */
	0,0,	0,0,	0,0,	0,0,	0,0,	15,0,	0,0,	0,0,  /*  |   combinacao       */
	0,0,	16,0,	0,0,	0,0,	0,0,	0,0,	0,0,	0,0,  /*  |		       */
	0,0,	0,0,	0,0,	0,0,	0,0,	0,0,	0,0,	16,0, /*  |		       */
	0,0,	0,0,	0,0,	16,0,	0,0,	0,0,	0,0,	0,0,  /* -		       */
	0,0,	0,0,	7,'/',	0,0,	6,'?',	0,0,	5,2,	0,0,  /* -		       */
	8,'.',	0,0,	0,0,	0,0,	10,',', 0,0,	9,3,	0,0,  /*  |-> segundos estados */
	11,'-', 0,0,	13,BS,	0,0,	0,0,	0,0,	12,4,	0,0   /* -    de SHIFT	       */
	};

	struct transicao tr;
	static byte estado[ NUMTERM ] = { 0 };
	byte classe; /* classe do evento recebido */
	char codigo;

	if (ev.make) {
		switch (ev.tecla) {
			case '\n': { classe=0; break; } 	/* \n = 10   => tecla [*] */
			case '\013': { classe=4; break; }	/* \013 = 11 => tecla [#] */
			case '\0':  { classe=2; break; }	/* \0 = 0    => tecla [0] */
			default: { classe=6; }
		}
	}
	else  {
		switch (ev.tecla) {
			case '\n': { classe=1; break; } 	/* \n = 10   => tecla [*] */
			case '\013': { classe=5; break; }	/* \013 = 11 => tecla [#] */
			case '\0':  { classe=3; break; }	/* \0 = 0    => tecla [0] */
			default: { classe=7; }
		}
	}
	tr = transicoes[ estado[ term ] ] [ classe ];
	estado[ term ] = tr.ps; /* proximo estado */

	if (tr.ascii >= 1 && tr.ascii <= 4)
		codigo = tascii[ tr.ascii-1 ] [ ev.tecla-1 ];
	else
		codigo = tr.ascii;

	if (codigo!=0) {
		/* bufferiza tecla para mgetkey() */
		teclas[term][tec_ent[term]++] = codigo;
		tec_ent[term]&=0x0f; /* mascara corresp. a 16 */
		_signal( &skey[term] );
	}
}

void liga_alarmes ( term,alarmes )
int term;
char *alarmes;
{
 unsigned int p=4;
 int t;
 if (term==32) return;
 for ( t=0 ; t<=2 ; t++ ) {
	  if ( alarmes[t] == 'D')
		 terminais[term].alarme = terminais[term].alarme | p;
	   else
	     if ( alarmes[t] == 'L' )
		terminais[term].alarme = terminais[term].alarme & ~p;
	 p=p>>1;
	 }
	 terminais[term].alarme |= 8; /* indica alarme pendente */
	terminais[term].pend++;
}

/*
**	rotina chamada pela Int2f para bufferizar uma tecla da console
*/
void put_sysch( c )
byte c;
{
	sys_buf_char = c;
	_signal(&syskey);
}

/*
**	retorna uma tecla ja processada
*/
byte mgetkey( )
{
	char c;
	if (TID==32) {
		wait(&syskey);
		return sys_buf_char;
	}
	if (TID>32) return 0;
	if (tec_ent[TID]==tec_sai[TID])
		wait( &skey[TID] );
	disable();
	c=teclas[TID][tec_sai[TID]++];
	tec_sai[TID]&=0x0f;
	enable();
	return c;
}

static void printerbyte( term,c )
byte term,c;
{
	disable();
	if (c=='\n') printerbyte( term, '\r' );
	if (((terminais[term].ps-terminais[term].pe)==1)
		||(terminais[term].ps==0 && (terminais[term].pe+1)==MAXFILA1) || c==0 ){
			wait (&sprt[term]);
	}
	terminais[term].prt_fila[ terminais[term].pe++ ] = c;
	terminais[term].pe&=(MAXFILA1-1);

	enable();
}

/* static void pputstr( s )
char *s;
{
	if (!*s) return;
	while( *s )
		printerbyte( TID,*s++ );
} */

/* static void pprintf( char *fmt, ...)
{
	va_list argptr;
	int num=0;
	char str[100];
	char *s=fmt;
	while( *s ){
		if (*s++=='%')
			num++;
	}
	va_start( argptr, num );
	vsprintf( str, fmt, argptr );
	pputstr( str );
	va_end( argptr );
} */

/*
**  Envia um caracter ao display do terminal
*/
static void displaybyte( term, c )
byte term,c;
{
	disable();
	if (((terminais[term].s-terminais[term].e)==1)
		||(terminais[term].s==0 && terminais[term].e==MAXFILA))
			wait (&sdsp[term]);
	terminais[term].fila[ terminais[term].e++ ] = c;
	terminais[term].e&=(MAXFILA-1);
	enable();
}

/* constantes relacionadas com a programacao do display */
#define CLRSCR	1
#define LIN1	0x80
#define LIN2	0xc0
#define CURLEFT 04
#define CURRIG	06
#define CUROFF  0x0c
#define CURON   0x0e
#define CURDL   0x10
#define MAXCOLS 16

/*
**	limpa a tela do mt100
*/
static void mclrscr( )
{
	byte term=TID;
	displaybyte( term, 0x82  |  ( (CLRSCR & 0x80) >> 7 )  );
	displaybyte( term, CLRSCR & 0x7f  ); /* limpa d7 */
}

/*
**	posiciona o cursor em dada posicao
*/
static void mgotolc( lin, col )
byte lin,col;
{
	byte term=TID;
	int cont;
	if (!(col>0 && col<17) )
		col=1;
	if (lin==1)
		cont=LIN1;	/* endereco inicial no display */
	else
		cont=LIN2;	/* endereco inicial 2 linha	*/
	cont+=(col-1);		/* coluna de 1 a 16 */
	displaybyte( term, 0x82  |  ( (cont&0x80) >> 7 )  );
	displaybyte( term, cont & 0x7f	); /* limpa d7 */
}

/*
**	envia sequencia para "backspace" no display
*/
static void mcurback( )
{
	byte term=TID;
	/* DESLOCA PARA ESQUERDA */
	displaybyte( term, 0x82  |  ( (CURDL & 0x80) >> 7 )  );
	displaybyte( term, CURDL & 0x7f  ); /* limpa d7 */
	/* LIMPA CARACTER  */
	displaybyte( term, ' ' );
	/* DESLOCA PARA ESQUERDA */
	displaybyte( term, 0x82  |  ( (CURDL & 0x80) >> 7 )  );
	displaybyte( term, CURDL & 0x7f  ); /* limpa d7 */
}

static byte mcol[NUMTERM];
static byte mlin[NUMTERM];
/*
** armazena posicao corrente
*/
void mgetpos( pos )
byte *pos;
{
	byte term=TID;
	if (term>=32) {
		pos[0]=(byte)wherex(); pos[1]=(byte)wherey(); return;
	}
	*pos++=mcol[term];
	*pos=mlin[term];
}

/*
** reposiciona o cursor
*/
void msetpos( pos )
byte *pos;
{
	byte term=TID;
	if (term>=32) {
		gotoxy(pos[0],pos[1]); return;
	}
	mcol[term]=*pos++;
	mlin[term]=*pos;
	mgotolc((*pos)+1,*(pos-1)+1);
}

/*
**	envia um caracter para o display, inclusive especiais
*/
void mputc( c,wait )
byte c,wait;
{
	byte term=TID;
	if (term>=32) {
		if (c=='\f') clrscr();
		else
			putchar(c);
		return;
	}
	if (c=='\f') {
		mclrscr();
		mlin[term]=mcol[term]=0;
	}
	else if (mlin[term]==2) {
		if (wait) {
			ALARMES("DLD");
			mgetkey();
		}
		mlin[term]=mcol[term]=0;
		mclrscr();
		mputc(c,wait);
	}
	else if ( c & 0x80 ){ 	/* checa se d7=1	*/
		displaybyte( term,c & 0x7f);
		mcol[term]++;
		if (mcol[term]>=MAXCOLS)
			mputc( '\n',wait );
	}
	else if (c>=' ') {
		displaybyte( term,c); /* limpa d7 */
		mcol[term]++;
		if (mcol[term]>=MAXCOLS)
			mputc( '\n',wait );
	}
	else
		switch (c) {
		case '\n': /* avanca uma linha */
			if (mlin[term]==1) {
				/* linha estourada, aguarde tecla */
				mlin[term]=2; mcol[term]=0;
			}
			else
			{
				mlin[term]++; mcol[term]=0; mgotolc(2,1);
			}
			break;
		case '\r': /* retorna para o inicio desta linha */
			mgotolc( mlin[term]+1, 1 );
			mcol[term]=0;
			break;
		case '\b': /* backspace */
			if (mcol[term]>0) {
				mcurback();
				mcol[term]--;
			}
			break;
		default:
			displaybyte( term,c );
			mcol[term]++;
			if (mcol[term]>=MAXCOLS)
				mputc( '\n',wait );
			break;
		}
}

/*
**	envia uma sequencia de caracteres para o display
*/
void mputstr( s )
char *s;
{
	if (!*s)
		return;
	while( *s ){
		mputc( *s++,1 );
	}
}

/*
** mostra string com formatacao no display
*/
/*void mprintf( fmt, ...)
char *fmt;
{
	va_list argptr;
	char str[129];
	va_start( argptr, fmt );
	vsprintf( str, fmt, argptr );
	if (TID>=32)
		puts(str);
	else
		mputstr( str );
	va_end( argptr );
}*/

/* envia um beep para o microterminal
void mbeep( dur,nota )
byte dur,nota;
{
	byte term=TID;
	displaybyte( term, 0x80 + (dur<<5) + (nota<<2) );
} */

void goxy( col,lin ) 
int col,lin;
{
	byte term=TID;
	if (term>=32) {
		gotoxy(col,lin);
		return;
	}
	mcol[term]=(byte)(col-1);
	mlin[term]=(byte)(lin-1);
	mgotolc(lin,col);
}

PROCESS task[NUMTERM+5];	/* array de processos para os terminais */

static void inicia_processos( )
{
	int i;
	init_tbl=load_task_table();
	for (i=0;i<NUMTERM+5;i++)
		if ((init_tbl+i)->stack_words)
			start_task(
				task+i,
				(init_tbl+i)->stack_words,
				(init_tbl+i)->function,
				(init_tbl+i)->name,
				(init_tbl+i)->terminal
			);
	init_spooler();
}

void main( argc,argv )
int argc;
char **argv;
{
/*	int fp;
	void *p;
*/
	mtnet_init(argc, argv );	/* inicializacao */
	_getcwd(current_dir,64);
	strcat(current_dir,"\\");
/*printf("\nDiretorio corrente: %s\n",current_dir);*/
	inicia_processos();
/*printf("\nPSP corrente: %04x\n",_psp);*/
/*	fp=_creat("mttask.dat",0);
	p=(void *)init_tbl;
	_write(fp,(char *)&p,4);
	p=(void *)task;
	_write(fp,(char *)&p,4);
	_close(fp);
*/
	mtnet_install();
}

void stop_run( ) {
	ALARMES("DDD");
	mprintf("\f* TERMINAL OFF *");
	while(1) mgetkey();
}

void accept_inkey( buffer ) 
char *buffer;
{
	char c,s[4];
	if (tec_ent[TID]==tec_sai[TID])
		c=0;
	else {
		disable();
		c=teclas[TID][tec_sai[TID]++];
		tec_sai[TID]&=0x0f;
		enable();
	}
	sprintf(s,"%03.3d",c);
	move_bytes(buffer,s,3);
}

/**************************************************************************/
#endif

void runtime_err( ) { }

void init_terminal( ) {
	if (TID>32) return;
	if (TID<NUMTERM) ALARMES("LLD");
	while(mgetkey()!=DEL);
	mputc('\f',0);
}

void display( buffer,len ) 
char *buffer;
int len;
{ /* anteriormente disp_glob() */
	while (len--)
		mputc(*buffer++,1);
}

void newline( ) {
	if (TID<=32)
		mputc('\n',1);
}

void display_erase( ) {
	mputc('\f',1);
}

/* end of MCOBLIB.C */
