/*
**		Saida formatada para o MCobol
*/
#include                <stdio.h>
#include                <ctype.h>

#include		<stdarg.h>	/* modificado para uso de va_arg */

#define EOS             ('\0')
#define FALSE           0
#define TRUE            1
#define BITS_PER_CHAR   8

/*
 * WORKSIZE is the dimension of work[] which must hold enough characters
 * to store a long integer (expanded using the %b format) + '-' and EOS
 */
#define WORKSIZE        (((sizeof (long)) * BITS_PER_CHAR) + 2)

struct  fmtbuf {
        char    f_type;         /* Format type ('d', etc.)              */
        char    f_width;        /* Argument width (INT or LONG)         */
        char    f_radix;        /* Argument radix (8, 10, etc.)         */
};

#define INT     0
#define LONG    1
#define NEG     2               /* Signal for output of '-'             */

/*
 * formatinfo[] stores information about the "value" formats
 */
static struct fmtbuf formatinfo[] = {
        { 'b',  INT,     2 },
        { 'd',  INT,    10 },
        { 'o',  INT,     8 },
        { 'u',  INT,    10 },
        { 'x',  INT,    16 },
        { 'B',  LONG,    2 },
        { 'D',  LONG,   10 },
        { 'O',  LONG,    8 },
        { 'U',  LONG,   10 },
        { 'X',  LONG,   16 },
        { EOS,  0,       0 },
};


static unsigned int count;
static unsigned char _iostr=0;	/* 0 - files, 1 - strings, 2 - microterminal */
static char *sbuf=NULL;

void _start( flg,buf )
unsigned char flg;
void *buf;
{
	_iostr = flg;
	count = 0;
	if (flg==1)	/* _iostr on? */
		sbuf=buf;
}

void _putc( c,... )
int c;
{
	if (!_iostr) {
		FILE *fd;
		va_list argp;
		va_start(argp,c);
		fd=va_arg(argp,FILE *);
		putc(c,fd);
	}
	else if (_iostr==2) /* output to microterminal */
		mputc(c,1);
	else
		*sbuf++ = c;		
	count++;
}

static int doprnt(format,fildes,argp)
char            *format;                /* Format string                */
va_list		argp;			/* Argument vector pointer	*/
FILE		fildes;			/* File descriptor		*/
{
        register int            c;
        register union {
            struct fmtbuf       *fmt;
            char                *result;
        } p;            
        char            ljust;          /* TRUE for left-justification  */
        char            work[WORKSIZE]; /* Number buffer                */
        char            fill;           /* '0' or ' ' for fill          */
        int             prec;           /* Precision                    */
        int             slen;           /* Field length                 */
        int             width;          /* Argument width               */
        unsigned long   value;		/* current argument value 	*/
        int             temp;           /* Set if '-' needed.           */
        short           radix;          /* Conversion radix             */
	int		tmparg;		/* for temp use (modified R.P.)	*/

        while ((c = *format++) != EOS) {
            /*
             * Check for conversion specifications.
             */
            if (c == '%') {
                /*
                 * Check for various options.
                 */
                c = *format++;
                ljust = 0;                      /* No left adjustment   */
                fill = ' ';                     /* Fill with spaces     */
                if (c == '-') {                 /* %-   left justify    */
                    ljust++;
                    c = *format++;
                }
                if (c == '0') {                 /* %0d  zero fill arg.  */
                    fill = c;
                    c = *format++;
                }
                if (c == '?' || c == '*') {     /* %*   width from arg  */
                    width = va_arg(argp,int); /* *argp++; */
                    c = *format++;
                }
                else {                          /* %n   field width     */
                    width = 0;
                    while (isdigit(c)) {
                        width *= 10;
                        width += (c - '0');
                        c = *format++;
                    }
                }
                if (c == '.') {                 /* %n.n precision       */
                    c = *format++;              /* %n.* (from arg)      */
                    if (c == '?' || c == '*') {
                        prec = va_arg(argp,int); /* *argp++; */
                        c = *format++;
                    }
                    else {
                        prec = 0;
                        while (isdigit(c)) {
                            prec *= 10;
                            prec += (c - '0');
                            c = *format++;
                        }
                    }
                }
                else {
                    prec = 32767;               /* No prec. specified   */
                }
                /*
                 * Process conversion chars, understanding longs.
                 */
                if (c == 'l' || c == 'L') {
                    switch (c = *format++) {
                    case 'b':
                    case 'd':
                    case 'o':
                    case 'u':
                    case 'x':
                        c = toupper(c); /* Make %lb %B          */
                    case 'B':
                    case 'D':
                    case 'O':
                    case 'U':
                    case 'X':
                        break;
                    default:            /* Here on %Lfoo ==> %Ufoo      */
                        c = 'U';
                        format--;
                        break;
                    }
                }
                /*
                 * Search the numeric format structure for this conversion.
                 */
                for (p.fmt = formatinfo;
                        (p.fmt->f_type != c && p.fmt->f_type != EOS);
                        p.fmt++)
                    ;
                if (p.fmt->f_type != EOS) {
                    /*
                     * A numeric conversion was found.
                     * Get the value and expand it into the work area.
                     */
                    radix = p.fmt->f_radix;
                    temp = p.fmt->f_width;
                    p.result = &work[WORKSIZE];
                    *--p.result = EOS;          /* Terminate result     */
                    if (temp == INT) {
			tmparg = va_arg(argp,int);
                        if (c == 'd' && tmparg /* *argp */ < 0) {
                            value = -tmparg; /* -*argp++; */
                            temp = NEG;         /* Remember signal      */
                        }
                        else
                            value = (unsigned) tmparg; /* *argp++; */
                    }
                    else {
                        value = va_arg(argp,long); /* *((long *) argp)++; */
                    }
#ifdef  TESTING
                    fprintf(stderr,
                        "radix %d, value = %d (signed), %u (unsigned), 0x%x\n",
                        radix, value, value, value);
#endif
                    if (value == 0)
                        *--p.result = '0';
                    else {
                        /*
                         * Convert an unsigned non-zero number.
                         */
                        do {
                            c = value % radix;
                            *--p.result = c + ((c < 10) ? '0' : ('a' - 10));
                        } while ((value /= radix) != 0);
                        if (temp == NEG)
                            *--p.result = '-';
                    }
#ifdef  TESTING
                    fprintf(stderr, "result \"%s\"\n", p.result);
#endif
                }
                else { 
                    /*
                     * String or something
                     */         
                    switch (c) {
                    case 'q':                   /* Funny int value      */
                        /*
                         * Convert a word as a pair of octal bytes.
                         */
                        c = 16384;
                        p.result = work;
                        temp = va_arg(argp,int); /* *argp++; */
                        while (c > 0) {
                            *p.result++ = (temp / c) + '0';
                            temp %= c;
                            if (c == 256) {
                                c = 64;
                                *p.result++ = '.';
                            }
                            else
                                c >>= 3;
                        }
                        *p.result = EOS;
                        p.result = work;
                        break;

                    /* case 'r':                   Remote format        
                        argp = (int *) *argp;
                        format = (char *) *argp++;
                        continue;                  No print here        */

                    case 's':                   /* String               */
                        if ((p.result = va_arg(argp,char *))
				/* (char *) *argp++) */ == NULL)
                            p.result = "{NULL}";        /* Bug hack     */
                        break;

                    case 'c':                   /* Character            */
                        c = va_arg(argp,char); /* *argp++; */
                        /*
                         * Fall through
                         */
                    default:                    /* %% or whatever       */
                        p.result = work;
                        *p.result = c;
                        work[1] = EOS;
                        break;
                    }
                }
                /*
                 * p.result -> first byte of string to output.
                 * Reuse c as a register temp.
                 */
                c = strlen(p.result);           /* True result length   */
                slen = (c > prec) ? prec : c;   /* Field length         */
#ifdef  TESTING
                fprintf(stderr, "width %d, field length %d, precision %d\n",
                    width, slen, prec);
#endif
                if (!ljust) {                   /* Right justify?       */
                    while (--width >= slen) {
                        _putc(fill, fildes);
                    }
                }
                /*
                 * Output the string (up to "prec" bytes)
                 */
                for (c = prec; *p.result != EOS && --c >= 0;) {
                    _putc(*p.result++, fildes);
                }
                if (ljust) {                    /* Left justify?        */
                    while (--width >= slen) {
                        _putc(' ', fildes);
                    }
                }
            }
            else {
                /*
                 * Not in a % thing, just output the byte.
                 */
                _putc(c, fildes);
            }
        }
	if (_iostr==1) 
		_putc(EOS,fildes);
/*        if ((fildes->_flag & _IOSTRG) != 0)
            putc(EOS, fildes);
*/
	return count;
}

int sprintf( buf,format,... )
char *buf;
char *format;
{
	va_list argp;
	va_start(argp,format);
	_start(1,buf);
	return doprnt(format,NULL,argp);
}

int fprintf( fp,format,... )
FILE *fp;
char *format;
{
	va_list argp;
	va_start(argp,format);
	_start(0,NULL);
	return doprnt(format,fp,argp);
}

int printf( format,... )
char *format;
{
	va_list argp;
	va_start(argp,format);
	_start(0,NULL);
	return doprnt(format,stdout,argp);
}

int mprintf( format,... )
char *format;
{
	va_list argp;
	va_start(argp,format);
	_start(2,NULL);
	return doprnt(format,NULL,argp);
}

int vsprintf( buf,format,argp )
char *buf;
char *format;
va_list argp;
{
	_start(1,buf);
	return doprnt(format,NULL,argp);
}

int vfprintf( fp,format,argp )
FILE *fp;
char *format;
va_list argp;
{
	_start(0,NULL);
	return doprnt(format,fp,argp);
}

int vprintf( format,argp )
char *format;
va_list argp;
{
	_start(0,NULL);
	return doprnt(format,stdout,argp);
}

/* end of MCPRINTF.C */
